# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import TYPE_CHECKING, List, Optional
from ..call import call, call_async, call_sync
from ..dto import requests as dto
from ..dto.ascii.response import Response
from ..ascii import Axis, AxisSettings, AxisStorage, Warnings
from ..dto.ascii.set_state_axis_response import SetStateAxisResponse

if TYPE_CHECKING:
    from .illuminator import Illuminator


class IlluminatorChannel:
    """
    Use to control a channel (LED lamp) on an illuminator.
    Requires at least Firmware 7.09.
    """

    @property
    def illuminator(self) -> 'Illuminator':
        """
        Illuminator of this channel.
        """
        return self._illuminator

    @property
    def channel_number(self) -> int:
        """
        The channel number identifies the channel on the illuminator.
        """
        return self._channel_number

    @property
    def settings(self) -> AxisSettings:
        """
        Settings and properties of this channel.
        """
        return self._settings

    @property
    def storage(self) -> AxisStorage:
        """
        Key-value storage of this channel.
        """
        return self._storage

    @property
    def warnings(self) -> Warnings:
        """
        Warnings and faults of this channel.
        """
        return self._warnings

    def __init__(self, illuminator: 'Illuminator', channel_number: int):
        self._illuminator = illuminator
        self._channel_number = channel_number
        self._axis = Axis(illuminator.device, channel_number)
        self._settings = AxisSettings(self._axis)
        self._storage = AxisStorage(self._axis)
        self._warnings = Warnings(illuminator.device, channel_number)

    def on(
            self
    ) -> None:
        """
        Turns this channel on.
        """
        request = dto.ChannelOn(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            on=True,
        )
        call("illuminator/on", request)

    async def on_async(
            self
    ) -> None:
        """
        Turns this channel on.
        """
        request = dto.ChannelOn(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            on=True,
        )
        await call_async("illuminator/on", request)

    def off(
            self
    ) -> None:
        """
        Turns this channel off.
        """
        request = dto.ChannelOn(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            on=False,
        )
        call("illuminator/on", request)

    async def off_async(
            self
    ) -> None:
        """
        Turns this channel off.
        """
        request = dto.ChannelOn(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            on=False,
        )
        await call_async("illuminator/on", request)

    def set_on(
            self,
            on: bool
    ) -> None:
        """
        Turns this channel on or off.

        Args:
            on: True to turn channel on, false to turn it off.
        """
        request = dto.ChannelOn(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            on=on,
        )
        call("illuminator/on", request)

    async def set_on_async(
            self,
            on: bool
    ) -> None:
        """
        Turns this channel on or off.

        Args:
            on: True to turn channel on, false to turn it off.
        """
        request = dto.ChannelOn(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            on=on,
        )
        await call_async("illuminator/on", request)

    def is_on(
            self
    ) -> bool:
        """
        Checks if this channel is on.

        Returns:
            True if channel is on, false otherwise.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
        )
        response = call(
            "illuminator/is_on",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    async def is_on_async(
            self
    ) -> bool:
        """
        Checks if this channel is on.

        Returns:
            True if channel is on, false otherwise.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
        )
        response = await call_async(
            "illuminator/is_on",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    def set_intensity(
            self,
            intensity: float
    ) -> None:
        """
        Sets channel intensity as a fraction of the maximum flux.

        Args:
            intensity: Fraction of intensity to set (between 0 and 1).
        """
        request = dto.ChannelSetIntensity(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            intensity=intensity,
        )
        call("illuminator/set_intensity", request)

    async def set_intensity_async(
            self,
            intensity: float
    ) -> None:
        """
        Sets channel intensity as a fraction of the maximum flux.

        Args:
            intensity: Fraction of intensity to set (between 0 and 1).
        """
        request = dto.ChannelSetIntensity(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            intensity=intensity,
        )
        await call_async("illuminator/set_intensity", request)

    def get_intensity(
            self
    ) -> float:
        """
        Gets the current intensity of this channel.

        Returns:
            Current intensity as fraction of maximum flux.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
        )
        response = call(
            "illuminator/get_intensity",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    async def get_intensity_async(
            self
    ) -> float:
        """
        Gets the current intensity of this channel.

        Returns:
            Current intensity as fraction of maximum flux.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
        )
        response = await call_async(
            "illuminator/get_intensity",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def generic_command(
            self,
            command: str,
            check_errors: bool = True,
            timeout: int = 0
    ) -> Response:
        """
        Sends a generic ASCII command to this channel.
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            check_errors: Controls whether to throw an exception when the device rejects the command.
            timeout: The timeout, in milliseconds, for a device to respond to the command.
                Overrides the connection default request timeout.

        Returns:
            A response to the command.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            command=command,
            check_errors=check_errors,
            timeout=timeout,
        )
        response = call(
            "interface/generic_command",
            request,
            Response.from_binary)
        return response

    async def generic_command_async(
            self,
            command: str,
            check_errors: bool = True,
            timeout: int = 0
    ) -> Response:
        """
        Sends a generic ASCII command to this channel.
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            check_errors: Controls whether to throw an exception when the device rejects the command.
            timeout: The timeout, in milliseconds, for a device to respond to the command.
                Overrides the connection default request timeout.

        Returns:
            A response to the command.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            command=command,
            check_errors=check_errors,
            timeout=timeout,
        )
        response = await call_async(
            "interface/generic_command",
            request,
            Response.from_binary)
        return response

    def generic_command_multi_response(
            self,
            command: str,
            check_errors: bool = True,
            timeout: int = 0
    ) -> List[Response]:
        """
        Sends a generic ASCII command to this channel and expects multiple responses.
        Responses are returned in order of arrival.
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            check_errors: Controls whether to throw an exception when a device rejects the command.
            timeout: The timeout, in milliseconds, for a device to respond to the command.
                Overrides the connection default request timeout.

        Returns:
            All responses to the command.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            command=command,
            check_errors=check_errors,
            timeout=timeout,
        )
        response = call(
            "interface/generic_command_multi_response",
            request,
            dto.GenericCommandResponseCollection.from_binary)
        return response.responses

    async def generic_command_multi_response_async(
            self,
            command: str,
            check_errors: bool = True,
            timeout: int = 0
    ) -> List[Response]:
        """
        Sends a generic ASCII command to this channel and expects multiple responses.
        Responses are returned in order of arrival.
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            check_errors: Controls whether to throw an exception when a device rejects the command.
            timeout: The timeout, in milliseconds, for a device to respond to the command.
                Overrides the connection default request timeout.

        Returns:
            All responses to the command.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            command=command,
            check_errors=check_errors,
            timeout=timeout,
        )
        response = await call_async(
            "interface/generic_command_multi_response",
            request,
            dto.GenericCommandResponseCollection.from_binary)
        return response.responses

    def generic_command_no_response(
            self,
            command: str
    ) -> None:
        """
        Sends a generic ASCII command to this channel without expecting a response and without adding a message ID
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            command=command,
        )
        call("interface/generic_command_no_response", request)

    async def generic_command_no_response_async(
            self,
            command: str
    ) -> None:
        """
        Sends a generic ASCII command to this channel without expecting a response and without adding a message ID
        For more information refer to: [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            command=command,
        )
        await call_async("interface/generic_command_no_response", request)

    def get_state(
            self
    ) -> str:
        """
        Returns a serialization of the current channel state that can be saved and reapplied.

        Returns:
            A serialization of the current state of the channel.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
        )
        response = call(
            "device/get_state",
            request,
            dto.StringResponse.from_binary)
        return response.value

    async def get_state_async(
            self
    ) -> str:
        """
        Returns a serialization of the current channel state that can be saved and reapplied.

        Returns:
            A serialization of the current state of the channel.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
        )
        response = await call_async(
            "device/get_state",
            request,
            dto.StringResponse.from_binary)
        return response.value

    def set_state(
            self,
            state: str
    ) -> SetStateAxisResponse:
        """
        Applies a saved state to this channel.

        Args:
            state: The state object to apply to this channel.

        Returns:
            Reports of any issues that were handled, but caused the state to not be exactly restored.
        """
        request = dto.SetStateRequest(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            state=state,
        )
        response = call(
            "device/set_axis_state",
            request,
            SetStateAxisResponse.from_binary)
        return response

    async def set_state_async(
            self,
            state: str
    ) -> SetStateAxisResponse:
        """
        Applies a saved state to this channel.

        Args:
            state: The state object to apply to this channel.

        Returns:
            Reports of any issues that were handled, but caused the state to not be exactly restored.
        """
        request = dto.SetStateRequest(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            state=state,
        )
        response = await call_async(
            "device/set_axis_state",
            request,
            SetStateAxisResponse.from_binary)
        return response

    def can_set_state(
            self,
            state: str
    ) -> Optional[str]:
        """
        Checks if a state can be applied to this channel.
        This only covers exceptions that can be determined statically such as mismatches of ID or version,
        the process of applying the state can still fail when running.

        Args:
            state: The state object to check against.

        Returns:
            An explanation of why this state cannot be set to this channel.
        """
        request = dto.CanSetStateRequest(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            state=state,
        )
        response = call(
            "device/can_set_axis_state",
            request,
            dto.CanSetStateAxisResponse.from_binary)
        return response.error

    async def can_set_state_async(
            self,
            state: str
    ) -> Optional[str]:
        """
        Checks if a state can be applied to this channel.
        This only covers exceptions that can be determined statically such as mismatches of ID or version,
        the process of applying the state can still fail when running.

        Args:
            state: The state object to check against.

        Returns:
            An explanation of why this state cannot be set to this channel.
        """
        request = dto.CanSetStateRequest(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            state=state,
        )
        response = await call_async(
            "device/can_set_axis_state",
            request,
            dto.CanSetStateAxisResponse.from_binary)
        return response.error

    def __repr__(
            self
    ) -> str:
        """
        Returns a string that represents the channel.

        Returns:
            A string that represents the channel.
        """
        request = dto.AxisToStringRequest(
            interface_id=self.illuminator.device.connection.interface_id,
            device=self.illuminator.device.device_address,
            axis=self.channel_number,
            type_override="Channel",
        )
        response = call_sync(
            "device/axis_to_string",
            request,
            dto.StringResponse.from_binary)
        return response.value



# pylint: disable=line-too-long
# The following imports are automatically generated by templates.
from .autofocus import Autofocus as Autofocus
from .camera_trigger import CameraTrigger as CameraTrigger
from .filter_changer import FilterChanger as FilterChanger
from .illuminator import Illuminator as Illuminator
from .illuminator_channel import IlluminatorChannel as IlluminatorChannel
from .microscope import Microscope as Microscope
from .objective_changer import ObjectiveChanger as ObjectiveChanger
from .wdi_autofocus_provider import WdiAutofocusProvider as WdiAutofocusProvider, AsyncWdiAutofocusOpener as AsyncWdiAutofocusOpener
from ..dto.microscopy.autofocus_status import AutofocusStatus as AutofocusStatus
from ..dto.microscopy.microscope_config import MicroscopeConfig as MicroscopeConfig
from ..dto.microscopy.third_party_components import ThirdPartyComponents as ThirdPartyComponents
from ..dto.microscopy.wdi_autofocus_provider_status import WdiAutofocusProviderStatus as WdiAutofocusProviderStatus



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson


@dataclass
class TranslateMessage:
    """
    Represents a message from translator regarding a block translation.
    """

    message: str
    """
    The message describing an occurrence.
    """

    from_block: int
    """
    The index in the block string that the message regards to.
    """

    to_block: int
    """
    The end index in the block string that the message regards to.
    The end index is exclusive.
    """

    @staticmethod
    def zero_values() -> 'TranslateMessage':
        return TranslateMessage(
            message="",
            from_block=0,
            to_block=0,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'TranslateMessage':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return TranslateMessage.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'message': str(self.message or ''),
            'fromBlock': int(self.from_block),
            'toBlock': int(self.to_block),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'TranslateMessage':
        return TranslateMessage(
            message=data.get('message'),  # type: ignore
            from_block=data.get('fromBlock'),  # type: ignore
            to_block=data.get('toBlock'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.message is not None:
            if not isinstance(self.message, str):
                raise ValueError(f'Property "Message" of "TranslateMessage" is not a string.')

        if self.from_block is None:
            raise ValueError(f'Property "FromBlock" of "TranslateMessage" is None.')

        if not isinstance(self.from_block, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "FromBlock" of "TranslateMessage" is not a number.')

        if int(self.from_block) != self.from_block:
            raise ValueError(f'Property "FromBlock" of "TranslateMessage" is not integer value.')

        if self.to_block is None:
            raise ValueError(f'Property "ToBlock" of "TranslateMessage" is None.')

        if not isinstance(self.to_block, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "ToBlock" of "TranslateMessage" is not a number.')

        if int(self.to_block) != self.to_block:
            raise ValueError(f'Property "ToBlock" of "TranslateMessage" is not integer value.')



# This file is generated. Do not modify by hand.
from enum import Enum


class LogOutputMode(Enum):
    """
    Mode of logging output of the library.
    """

    OFF = 0
    STDOUT = 1
    STDERR = 2
    FILE = 3



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, Optional
import decimal
import zaber_bson
from ..microscopy.third_party_components import ThirdPartyComponents


@dataclass
class MicroscopeFindRequest:

    interface_id: int = 0

    third_party: Optional[ThirdPartyComponents] = None

    @staticmethod
    def zero_values() -> 'MicroscopeFindRequest':
        return MicroscopeFindRequest(
            interface_id=0,
            third_party=None,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'MicroscopeFindRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return MicroscopeFindRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'interfaceId': int(self.interface_id),
            'thirdParty': self.third_party.to_dict() if self.third_party is not None else None,
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'MicroscopeFindRequest':
        return MicroscopeFindRequest(
            interface_id=data.get('interfaceId'),  # type: ignore
            third_party=ThirdPartyComponents.from_dict(data.get('thirdParty')) if data.get('thirdParty') is not None else None,  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.interface_id is None:
            raise ValueError(f'Property "InterfaceId" of "MicroscopeFindRequest" is None.')

        if not isinstance(self.interface_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "InterfaceId" of "MicroscopeFindRequest" is not a number.')

        if int(self.interface_id) != self.interface_id:
            raise ValueError(f'Property "InterfaceId" of "MicroscopeFindRequest" is not integer value.')

        if self.third_party is not None:
            if not isinstance(self.third_party, ThirdPartyComponents):
                raise ValueError(f'Property "ThirdParty" of "MicroscopeFindRequest" is not an instance of "ThirdPartyComponents".')

            self.third_party.validate()



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson
from .axis_move_type import AxisMoveType
from ...units import Units, UnitsAndLiterals, units_from_literals


@dataclass
class LockstepMoveRequest:

    interface_id: int = 0

    device: int = 0

    lockstep_group_id: int = 0

    wait_until_idle: bool = False

    type: AxisMoveType = next(first for first in AxisMoveType)

    arg: float = 0

    unit: UnitsAndLiterals = Units.NATIVE

    velocity: float = 0

    velocity_unit: UnitsAndLiterals = Units.NATIVE

    acceleration: float = 0

    acceleration_unit: UnitsAndLiterals = Units.NATIVE

    @staticmethod
    def zero_values() -> 'LockstepMoveRequest':
        return LockstepMoveRequest(
            interface_id=0,
            device=0,
            lockstep_group_id=0,
            wait_until_idle=False,
            type=next(first for first in AxisMoveType),
            arg=0,
            unit=Units.NATIVE,
            velocity=0,
            velocity_unit=Units.NATIVE,
            acceleration=0,
            acceleration_unit=Units.NATIVE,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'LockstepMoveRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return LockstepMoveRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'interfaceId': int(self.interface_id),
            'device': int(self.device),
            'lockstepGroupId': int(self.lockstep_group_id),
            'waitUntilIdle': bool(self.wait_until_idle),
            'type': self.type.value,
            'arg': float(self.arg),
            'unit': units_from_literals(self.unit).value,
            'velocity': float(self.velocity),
            'velocityUnit': units_from_literals(self.velocity_unit).value,
            'acceleration': float(self.acceleration),
            'accelerationUnit': units_from_literals(self.acceleration_unit).value,
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'LockstepMoveRequest':
        return LockstepMoveRequest(
            interface_id=data.get('interfaceId'),  # type: ignore
            device=data.get('device'),  # type: ignore
            lockstep_group_id=data.get('lockstepGroupId'),  # type: ignore
            wait_until_idle=data.get('waitUntilIdle'),  # type: ignore
            type=AxisMoveType(data.get('type')),  # type: ignore
            arg=data.get('arg'),  # type: ignore
            unit=Units(data.get('unit')),  # type: ignore
            velocity=data.get('velocity'),  # type: ignore
            velocity_unit=Units(data.get('velocityUnit')),  # type: ignore
            acceleration=data.get('acceleration'),  # type: ignore
            acceleration_unit=Units(data.get('accelerationUnit')),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.interface_id is None:
            raise ValueError(f'Property "InterfaceId" of "LockstepMoveRequest" is None.')

        if not isinstance(self.interface_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "InterfaceId" of "LockstepMoveRequest" is not a number.')

        if int(self.interface_id) != self.interface_id:
            raise ValueError(f'Property "InterfaceId" of "LockstepMoveRequest" is not integer value.')

        if self.device is None:
            raise ValueError(f'Property "Device" of "LockstepMoveRequest" is None.')

        if not isinstance(self.device, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Device" of "LockstepMoveRequest" is not a number.')

        if int(self.device) != self.device:
            raise ValueError(f'Property "Device" of "LockstepMoveRequest" is not integer value.')

        if self.lockstep_group_id is None:
            raise ValueError(f'Property "LockstepGroupId" of "LockstepMoveRequest" is None.')

        if not isinstance(self.lockstep_group_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "LockstepGroupId" of "LockstepMoveRequest" is not a number.')

        if int(self.lockstep_group_id) != self.lockstep_group_id:
            raise ValueError(f'Property "LockstepGroupId" of "LockstepMoveRequest" is not integer value.')

        if self.type is None:
            raise ValueError(f'Property "Type" of "LockstepMoveRequest" is None.')

        if not isinstance(self.type, AxisMoveType):
            raise ValueError(f'Property "Type" of "LockstepMoveRequest" is not an instance of "AxisMoveType".')

        if self.arg is None:
            raise ValueError(f'Property "Arg" of "LockstepMoveRequest" is None.')

        if not isinstance(self.arg, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Arg" of "LockstepMoveRequest" is not a number.')

        if self.unit is None:
            raise ValueError(f'Property "Unit" of "LockstepMoveRequest" is None.')

        if not isinstance(self.unit, (Units, str)):
            raise ValueError(f'Property "Unit" of "LockstepMoveRequest" is not Units.')

        if self.velocity is None:
            raise ValueError(f'Property "Velocity" of "LockstepMoveRequest" is None.')

        if not isinstance(self.velocity, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Velocity" of "LockstepMoveRequest" is not a number.')

        if self.velocity_unit is None:
            raise ValueError(f'Property "VelocityUnit" of "LockstepMoveRequest" is None.')

        if not isinstance(self.velocity_unit, (Units, str)):
            raise ValueError(f'Property "VelocityUnit" of "LockstepMoveRequest" is not Units.')

        if self.acceleration is None:
            raise ValueError(f'Property "Acceleration" of "LockstepMoveRequest" is None.')

        if not isinstance(self.acceleration, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Acceleration" of "LockstepMoveRequest" is not a number.')

        if self.acceleration_unit is None:
            raise ValueError(f'Property "AccelerationUnit" of "LockstepMoveRequest" is None.')

        if not isinstance(self.acceleration_unit, (Units, str)):
            raise ValueError(f'Property "AccelerationUnit" of "LockstepMoveRequest" is not Units.')



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import zaber_bson


@dataclass
class AutofocusStatus:
    """
    Status of the autofocus.
    """

    in_focus: bool
    """
    Indicates whether the autofocus is in focus.
    """

    in_range: bool
    """
    Indicates whether the autofocus is in range.
    """

    @staticmethod
    def zero_values() -> 'AutofocusStatus':
        return AutofocusStatus(
            in_focus=False,
            in_range=False,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'AutofocusStatus':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return AutofocusStatus.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'inFocus': bool(self.in_focus),
            'inRange': bool(self.in_range),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'AutofocusStatus':
        return AutofocusStatus(
            in_focus=data.get('inFocus'),  # type: ignore
            in_range=data.get('inRange'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        pass



import asyncio
import threading
from typing import Coroutine, TypeVar, List, Any
from typing import Optional


class ThreadedEventLoop:
    """ Provides a thread-safe lazy-initialized asyncio event loop. """
    _loop: Optional[asyncio.AbstractEventLoop] = None
    _loop_lock = threading.Lock()
    _thread: Optional[threading.Thread] = None

    def _start_thread(self) -> None:
        self._loop = asyncio.new_event_loop()
        self._thread = threading.Thread(target=self._thread_loop)
        self._thread.daemon = True
        self._thread.name = "Zaber Motion Library Event Loop"
        self._thread.start()

    def _thread_loop(self) -> None:
        assert self._loop is not None
        asyncio.set_event_loop(self._loop)
        self._loop.run_forever()

    def get_loop(self) -> asyncio.AbstractEventLoop:
        with self._loop_lock:
            if self._loop is None:
                self._start_thread()
            assert self._loop is not None
            return self._loop


event_loop = ThreadedEventLoop()

TReturn = TypeVar('TReturn')


async def gather_all(*coro: Coroutine[Any, Any, TReturn]) -> List[TReturn]:
    """ Run all coroutines in parallel. """
    return await asyncio.gather(*coro)


def wait_all(*coro: Coroutine[Any, Any, TReturn]) -> List[TReturn]:
    """ Run all coroutines in parallel in thread pool executor and wait for them to finish. """
    future = asyncio.run_coroutine_threadsafe(gather_all(*coro), event_loop.get_loop())
    return future.result()



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson


@dataclass
class TranslatorTranslateRequest:

    translator_id: int = 0

    block: str = ""

    @staticmethod
    def zero_values() -> 'TranslatorTranslateRequest':
        return TranslatorTranslateRequest(
            translator_id=0,
            block="",
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'TranslatorTranslateRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return TranslatorTranslateRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'translatorId': int(self.translator_id),
            'block': str(self.block or ''),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'TranslatorTranslateRequest':
        return TranslatorTranslateRequest(
            translator_id=data.get('translatorId'),  # type: ignore
            block=data.get('block'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.translator_id is None:
            raise ValueError(f'Property "TranslatorId" of "TranslatorTranslateRequest" is None.')

        if not isinstance(self.translator_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "TranslatorId" of "TranslatorTranslateRequest" is not a number.')

        if int(self.translator_id) != self.translator_id:
            raise ValueError(f'Property "TranslatorId" of "TranslatorTranslateRequest" is not integer value.')

        if self.block is not None:
            if not isinstance(self.block, str):
                raise ValueError(f'Property "Block" of "TranslatorTranslateRequest" is not a string.')



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, Optional
import decimal
import zaber_bson
from .pvt_axis_type import PvtAxisType


@dataclass
class PvtAxisDefinition:
    """
    Defines an axis of the PVT sequence.
    """

    axis_number: int
    """
    Number of a physical axis or a lockstep group.
    """

    axis_type: Optional[PvtAxisType] = None
    """
    Defines the type of the axis.
    """

    @staticmethod
    def zero_values() -> 'PvtAxisDefinition':
        return PvtAxisDefinition(
            axis_number=0,
            axis_type=None,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'PvtAxisDefinition':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return PvtAxisDefinition.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'axisNumber': int(self.axis_number),
            'axisType': self.axis_type.value if self.axis_type is not None else None,
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'PvtAxisDefinition':
        return PvtAxisDefinition(
            axis_number=data.get('axisNumber'),  # type: ignore
            axis_type=PvtAxisType(data.get('axisType')) if data.get('axisType') is not None else None,  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.axis_number is None:
            raise ValueError(f'Property "AxisNumber" of "PvtAxisDefinition" is None.')

        if not isinstance(self.axis_number, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "AxisNumber" of "PvtAxisDefinition" is not a number.')

        if int(self.axis_number) != self.axis_number:
            raise ValueError(f'Property "AxisNumber" of "PvtAxisDefinition" is not integer value.')

        if self.axis_type is not None:
            if not isinstance(self.axis_type, PvtAxisType):
                raise ValueError(f'Property "AxisType" of "PvtAxisDefinition" is not an instance of "PvtAxisType".')



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from .command_failed_exception import CommandFailedException


class BadDataException(CommandFailedException):
    """
    Thrown when a parameter of a command is judged to be out of range by the receiving device.
    """



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, Optional
import decimal
import zaber_bson


@dataclass
class SimpleTuningParamDefinition:
    """
    Information about a parameter used for the simple tuning method.
    """

    name: str
    """
    The name of the parameter.
    """

    min_label: str
    """
    The human readable description of the effect of a lower value on this setting.
    """

    max_label: str
    """
    The human readable description of the effect of a higher value on this setting.
    """

    data_type: str
    """
    How this parameter will be parsed by the tuner.
    """

    default_value: Optional[float] = None
    """
    The default value of this parameter.
    """

    @staticmethod
    def zero_values() -> 'SimpleTuningParamDefinition':
        return SimpleTuningParamDefinition(
            name="",
            min_label="",
            max_label="",
            data_type="",
            default_value=None,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'SimpleTuningParamDefinition':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return SimpleTuningParamDefinition.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'name': str(self.name or ''),
            'minLabel': str(self.min_label or ''),
            'maxLabel': str(self.max_label or ''),
            'dataType': str(self.data_type or ''),
            'defaultValue': float(self.default_value) if self.default_value is not None else None,
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'SimpleTuningParamDefinition':
        return SimpleTuningParamDefinition(
            name=data.get('name'),  # type: ignore
            min_label=data.get('minLabel'),  # type: ignore
            max_label=data.get('maxLabel'),  # type: ignore
            data_type=data.get('dataType'),  # type: ignore
            default_value=data.get('defaultValue'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.name is not None:
            if not isinstance(self.name, str):
                raise ValueError(f'Property "Name" of "SimpleTuningParamDefinition" is not a string.')

        if self.min_label is not None:
            if not isinstance(self.min_label, str):
                raise ValueError(f'Property "MinLabel" of "SimpleTuningParamDefinition" is not a string.')

        if self.max_label is not None:
            if not isinstance(self.max_label, str):
                raise ValueError(f'Property "MaxLabel" of "SimpleTuningParamDefinition" is not a string.')

        if self.data_type is not None:
            if not isinstance(self.data_type, str):
                raise ValueError(f'Property "DataType" of "SimpleTuningParamDefinition" is not a string.')

        if self.default_value is not None:
            if not isinstance(self.default_value, (int, float, decimal.Decimal)):
                raise ValueError(f'Property "DefaultValue" of "SimpleTuningParamDefinition" is not a number.')



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson
from ..firmware_version import FirmwareVersion


@dataclass
class DeviceIdentity:
    """
    Representation of data gathered during device identification.
    """

    device_id: int
    """
    Unique ID of the device hardware.
    """

    serial_number: int
    """
    Serial number of the device.
    """

    name: str
    """
    Name of the product.
    """

    axis_count: int
    """
    Number of axes this device has.
    """

    firmware_version: FirmwareVersion
    """
    Version of the firmware.
    """

    is_modified: bool
    """
    The device has hardware modifications.
    """

    is_integrated: bool
    """
    The device is an integrated product.
    """

    @staticmethod
    def zero_values() -> 'DeviceIdentity':
        return DeviceIdentity(
            device_id=0,
            serial_number=0,
            name="",
            axis_count=0,
            firmware_version=FirmwareVersion.zero_values(),
            is_modified=False,
            is_integrated=False,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'DeviceIdentity':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return DeviceIdentity.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'deviceId': int(self.device_id),
            'serialNumber': int(self.serial_number),
            'name': str(self.name or ''),
            'axisCount': int(self.axis_count),
            'firmwareVersion': self.firmware_version.to_dict(),
            'isModified': bool(self.is_modified),
            'isIntegrated': bool(self.is_integrated),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'DeviceIdentity':
        return DeviceIdentity(
            device_id=data.get('deviceId'),  # type: ignore
            serial_number=data.get('serialNumber'),  # type: ignore
            name=data.get('name'),  # type: ignore
            axis_count=data.get('axisCount'),  # type: ignore
            firmware_version=FirmwareVersion.from_dict(data.get('firmwareVersion')),  # type: ignore
            is_modified=data.get('isModified'),  # type: ignore
            is_integrated=data.get('isIntegrated'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.device_id is None:
            raise ValueError(f'Property "DeviceId" of "DeviceIdentity" is None.')

        if not isinstance(self.device_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "DeviceId" of "DeviceIdentity" is not a number.')

        if int(self.device_id) != self.device_id:
            raise ValueError(f'Property "DeviceId" of "DeviceIdentity" is not integer value.')

        if self.serial_number is None:
            raise ValueError(f'Property "SerialNumber" of "DeviceIdentity" is None.')

        if not isinstance(self.serial_number, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "SerialNumber" of "DeviceIdentity" is not a number.')

        if int(self.serial_number) != self.serial_number:
            raise ValueError(f'Property "SerialNumber" of "DeviceIdentity" is not integer value.')

        if self.name is not None:
            if not isinstance(self.name, str):
                raise ValueError(f'Property "Name" of "DeviceIdentity" is not a string.')

        if self.axis_count is None:
            raise ValueError(f'Property "AxisCount" of "DeviceIdentity" is None.')

        if not isinstance(self.axis_count, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "AxisCount" of "DeviceIdentity" is not a number.')

        if int(self.axis_count) != self.axis_count:
            raise ValueError(f'Property "AxisCount" of "DeviceIdentity" is not integer value.')

        if self.firmware_version is None:
            raise ValueError(f'Property "FirmwareVersion" of "DeviceIdentity" is None.')

        if not isinstance(self.firmware_version, FirmwareVersion):
            raise ValueError(f'Property "FirmwareVersion" of "DeviceIdentity" is not an instance of "FirmwareVersion".')

        self.firmware_version.validate()



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long
from .alert_event import AlertEvent as AlertEvent
from .axis_identity import AxisIdentity as AxisIdentity
from .axis_type import AxisType as AxisType
from .can_set_state_axis_response import CanSetStateAxisResponse as CanSetStateAxisResponse
from .can_set_state_device_response import CanSetStateDeviceResponse as CanSetStateDeviceResponse
from .conversion_factor import ConversionFactor as ConversionFactor
from .device_identity import DeviceIdentity as DeviceIdentity
from .device_io_info import DeviceIOInfo as DeviceIOInfo
from .digital_output_action import DigitalOutputAction as DigitalOutputAction
from .get_axis_setting import GetAxisSetting as GetAxisSetting
from .get_axis_setting_result import GetAxisSettingResult as GetAxisSettingResult
from .get_setting import GetSetting as GetSetting
from .get_setting_result import GetSettingResult as GetSettingResult
from .io_port_label import IoPortLabel as IoPortLabel
from .io_port_type import IoPortType as IoPortType
from .lockstep_axes import LockstepAxes as LockstepAxes
from .message_type import MessageType as MessageType
from .oscilloscope_capture_properties import OscilloscopeCaptureProperties as OscilloscopeCaptureProperties
from .oscilloscope_data_source import OscilloscopeDataSource as OscilloscopeDataSource
from .paramset_info import ParamsetInfo as ParamsetInfo
from .pid_tuning import PidTuning as PidTuning
from .pvt_axis_definition import PvtAxisDefinition as PvtAxisDefinition
from .pvt_axis_type import PvtAxisType as PvtAxisType
from .pvt_mode import PvtMode as PvtMode
from .response import Response as Response
from .servo_tuning_param import ServoTuningParam as ServoTuningParam
from .servo_tuning_paramset import ServoTuningParamset as ServoTuningParamset
from .set_state_axis_response import SetStateAxisResponse as SetStateAxisResponse
from .set_state_device_response import SetStateDeviceResponse as SetStateDeviceResponse
from .simple_tuning import SimpleTuning as SimpleTuning
from .simple_tuning_param_definition import SimpleTuningParamDefinition as SimpleTuningParamDefinition
from .stream_axis_definition import StreamAxisDefinition as StreamAxisDefinition
from .stream_axis_type import StreamAxisType as StreamAxisType
from .stream_mode import StreamMode as StreamMode
from .trigger_action import TriggerAction as TriggerAction
from .trigger_condition import TriggerCondition as TriggerCondition
from .trigger_enabled_state import TriggerEnabledState as TriggerEnabledState
from .trigger_operation import TriggerOperation as TriggerOperation
from .trigger_state import TriggerState as TriggerState
from .unknown_response_event import UnknownResponseEvent as UnknownResponseEvent



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson
from ...units import Units, UnitsAndLiterals, units_from_literals


@dataclass
class ConversionFactor:
    """
    Represents unit conversion factor for a single dimension.
    """

    setting: str
    """
    Setting representing the dimension.
    """

    value: float
    """
    Value representing 1 native device unit in specified real-word units.
    """

    unit: UnitsAndLiterals
    """
    Units of the value.
    """

    @staticmethod
    def zero_values() -> 'ConversionFactor':
        return ConversionFactor(
            setting="",
            value=0,
            unit=Units.NATIVE,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'ConversionFactor':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return ConversionFactor.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'setting': str(self.setting or ''),
            'value': float(self.value),
            'unit': units_from_literals(self.unit).value,
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'ConversionFactor':
        return ConversionFactor(
            setting=data.get('setting'),  # type: ignore
            value=data.get('value'),  # type: ignore
            unit=Units(data.get('unit')),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.setting is not None:
            if not isinstance(self.setting, str):
                raise ValueError(f'Property "Setting" of "ConversionFactor" is not a string.')

        if self.value is None:
            raise ValueError(f'Property "Value" of "ConversionFactor" is None.')

        if not isinstance(self.value, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Value" of "ConversionFactor" is not a number.')

        if self.unit is None:
            raise ValueError(f'Property "Unit" of "ConversionFactor" is None.')

        if not isinstance(self.unit, (Units, str)):
            raise ValueError(f'Property "Unit" of "ConversionFactor" is not Units.')



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson


@dataclass
class CommandTooLongExceptionData:
    """
    Information describing why the command could not fit.
    """

    fit: str
    """
    The part of the command that could be successfully fit in the space provided by the protocol.
    """

    remainder: str
    """
    The part of the command that could not fit within the space provided.
    """

    packet_size: int
    """
    The length of the ascii string that can be written to a single line.
    """

    packets_max: int
    """
    The number of lines a command can be split over using continuations.
    """

    @staticmethod
    def zero_values() -> 'CommandTooLongExceptionData':
        return CommandTooLongExceptionData(
            fit="",
            remainder="",
            packet_size=0,
            packets_max=0,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'CommandTooLongExceptionData':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return CommandTooLongExceptionData.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'fit': str(self.fit or ''),
            'remainder': str(self.remainder or ''),
            'packetSize': int(self.packet_size),
            'packetsMax': int(self.packets_max),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'CommandTooLongExceptionData':
        return CommandTooLongExceptionData(
            fit=data.get('fit'),  # type: ignore
            remainder=data.get('remainder'),  # type: ignore
            packet_size=data.get('packetSize'),  # type: ignore
            packets_max=data.get('packetsMax'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.fit is not None:
            if not isinstance(self.fit, str):
                raise ValueError(f'Property "Fit" of "CommandTooLongExceptionData" is not a string.')

        if self.remainder is not None:
            if not isinstance(self.remainder, str):
                raise ValueError(f'Property "Remainder" of "CommandTooLongExceptionData" is not a string.')

        if self.packet_size is None:
            raise ValueError(f'Property "PacketSize" of "CommandTooLongExceptionData" is None.')

        if not isinstance(self.packet_size, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "PacketSize" of "CommandTooLongExceptionData" is not a number.')

        if int(self.packet_size) != self.packet_size:
            raise ValueError(f'Property "PacketSize" of "CommandTooLongExceptionData" is not integer value.')

        if self.packets_max is None:
            raise ValueError(f'Property "PacketsMax" of "CommandTooLongExceptionData" is None.')

        if not isinstance(self.packets_max, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "PacketsMax" of "CommandTooLongExceptionData" is not a number.')

        if int(self.packets_max) != self.packets_max:
            raise ValueError(f'Property "PacketsMax" of "CommandTooLongExceptionData" is not integer value.')



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, List, Optional
from collections.abc import Iterable
import zaber_bson
from .axis_mapping import AxisMapping
from .axis_transformation import AxisTransformation


@dataclass
class TranslatorConfig:
    """
    Configuration of a translator.
    """

    axis_mappings: Optional[List[AxisMapping]] = None
    """
    Optional custom mapping of translator axes to stream axes.
    """

    axis_transformations: Optional[List[AxisTransformation]] = None
    """
    Optional transformation of axes.
    """

    @staticmethod
    def zero_values() -> 'TranslatorConfig':
        return TranslatorConfig(
            axis_mappings=None,
            axis_transformations=None,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'TranslatorConfig':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return TranslatorConfig.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'axisMappings': [item.to_dict() for item in self.axis_mappings] if self.axis_mappings is not None else [],
            'axisTransformations': [item.to_dict() for item in self.axis_transformations] if self.axis_transformations is not None else [],
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'TranslatorConfig':
        return TranslatorConfig(
            axis_mappings=[AxisMapping.from_dict(item) for item in data.get('axisMappings')] if data.get('axisMappings') is not None else None,  # type: ignore
            axis_transformations=[AxisTransformation.from_dict(item) for item in data.get('axisTransformations')] if data.get('axisTransformations') is not None else None,  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.axis_mappings is not None:
            if not isinstance(self.axis_mappings, Iterable):
                raise ValueError('Property "AxisMappings" of "TranslatorConfig" is not iterable.')

            for i, axis_mappings_item in enumerate(self.axis_mappings):
                if axis_mappings_item is None:
                    raise ValueError(f'Item {i} in property "AxisMappings" of "TranslatorConfig" is None.')

                if not isinstance(axis_mappings_item, AxisMapping):
                    raise ValueError(f'Item {i} in property "AxisMappings" of "TranslatorConfig" is not an instance of "AxisMapping".')

                axis_mappings_item.validate()

        if self.axis_transformations is not None:
            if not isinstance(self.axis_transformations, Iterable):
                raise ValueError('Property "AxisTransformations" of "TranslatorConfig" is not iterable.')

            for i, axis_transformations_item in enumerate(self.axis_transformations):
                if axis_transformations_item is None:
                    raise ValueError(f'Item {i} in property "AxisTransformations" of "TranslatorConfig" is None.')

                if not isinstance(axis_transformations_item, AxisTransformation):
                    raise ValueError(f'Item {i} in property "AxisTransformations" of "TranslatorConfig" is not an instance of "AxisTransformation".')

                axis_transformations_item.validate()



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #
from typing import List, Optional
from ..dto import requests as dto
from ..call import call, call_async
from .axis import Axis
from ..dto.ascii.servo_tuning_paramset import ServoTuningParamset
from ..dto.ascii.paramset_info import ParamsetInfo
from ..dto.ascii.servo_tuning_param import ServoTuningParam
from ..dto.ascii.simple_tuning import SimpleTuning
from ..dto.ascii.simple_tuning_param_definition import SimpleTuningParamDefinition
from ..dto.ascii.pid_tuning import PidTuning


class ServoTuner:
    """
    Exposes the capabilities to inspect and edit an axis' servo tuning.
    Requires at least Firmware 6.25 or 7.00.
    """

    @property
    def axis(self) -> Axis:
        """
        The axis that will be tuned.
        """
        return self._axis

    def __init__(self, axis: Axis):
        """
        Creates instance of ServoTuner for the given axis.
        """
        self._axis = axis

    def get_startup_paramset(
            self
    ) -> ServoTuningParamset:
        """
        Get the paramset that this device uses by default when it starts up.

        Returns:
            The paramset used when the device restarts.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
        )
        response = call(
            "servotuning/get_startup_set",
            request,
            dto.ServoTuningParamsetResponse.from_binary)
        return response.paramset

    async def get_startup_paramset_async(
            self
    ) -> ServoTuningParamset:
        """
        Get the paramset that this device uses by default when it starts up.

        Returns:
            The paramset used when the device restarts.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
        )
        response = await call_async(
            "servotuning/get_startup_set",
            request,
            dto.ServoTuningParamsetResponse.from_binary)
        return response.paramset

    def set_startup_paramset(
            self,
            paramset: ServoTuningParamset
    ) -> None:
        """
        Set the paramset that this device uses by default when it starts up.

        Args:
            paramset: The paramset to use at startup.
        """
        request = dto.ServoTuningRequest(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
            paramset=paramset,
        )
        call("servotuning/set_startup_set", request)

    async def set_startup_paramset_async(
            self,
            paramset: ServoTuningParamset
    ) -> None:
        """
        Set the paramset that this device uses by default when it starts up.

        Args:
            paramset: The paramset to use at startup.
        """
        request = dto.ServoTuningRequest(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
            paramset=paramset,
        )
        await call_async("servotuning/set_startup_set", request)

    def load_paramset(
            self,
            to_paramset: ServoTuningParamset,
            from_paramset: ServoTuningParamset
    ) -> None:
        """
        Load the values from one paramset into another.

        Args:
            to_paramset: The paramset to load into.
            from_paramset: The paramset to load from.
        """
        request = dto.LoadParamset(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
            to_paramset=to_paramset,
            from_paramset=from_paramset,
        )
        call("servotuning/load_paramset", request)

    async def load_paramset_async(
            self,
            to_paramset: ServoTuningParamset,
            from_paramset: ServoTuningParamset
    ) -> None:
        """
        Load the values from one paramset into another.

        Args:
            to_paramset: The paramset to load into.
            from_paramset: The paramset to load from.
        """
        request = dto.LoadParamset(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
            to_paramset=to_paramset,
            from_paramset=from_paramset,
        )
        await call_async("servotuning/load_paramset", request)

    def get_tuning(
            self,
            paramset: ServoTuningParamset
    ) -> ParamsetInfo:
        """
        Get the full set of tuning parameters used by the firmware driving this axis.

        Args:
            paramset: The paramset to get tuning for.

        Returns:
            The raw representation of the current tuning.
        """
        request = dto.ServoTuningRequest(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
            paramset=paramset,
        )
        response = call(
            "servotuning/get_raw",
            request,
            ParamsetInfo.from_binary)
        return response

    async def get_tuning_async(
            self,
            paramset: ServoTuningParamset
    ) -> ParamsetInfo:
        """
        Get the full set of tuning parameters used by the firmware driving this axis.

        Args:
            paramset: The paramset to get tuning for.

        Returns:
            The raw representation of the current tuning.
        """
        request = dto.ServoTuningRequest(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
            paramset=paramset,
        )
        response = await call_async(
            "servotuning/get_raw",
            request,
            ParamsetInfo.from_binary)
        return response

    def set_tuning(
            self,
            paramset: ServoTuningParamset,
            tuning_params: List[ServoTuningParam],
            set_unspecified_to_default: bool = False
    ) -> None:
        """
        Set individual tuning parameters.
        Only use this method if you have a strong understanding of Zaber specific tuning parameters.

        Args:
            paramset: The paramset to set tuning of.
            tuning_params: The params to set.
            set_unspecified_to_default: If true, any tuning parameters not included in TuningParams
                are reset to their default values.
        """
        request = dto.SetServoTuningRequest(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
            paramset=paramset,
            tuning_params=tuning_params,
            set_unspecified_to_default=set_unspecified_to_default,
        )
        call("servotuning/set_raw", request)

    async def set_tuning_async(
            self,
            paramset: ServoTuningParamset,
            tuning_params: List[ServoTuningParam],
            set_unspecified_to_default: bool = False
    ) -> None:
        """
        Set individual tuning parameters.
        Only use this method if you have a strong understanding of Zaber specific tuning parameters.

        Args:
            paramset: The paramset to set tuning of.
            tuning_params: The params to set.
            set_unspecified_to_default: If true, any tuning parameters not included in TuningParams
                are reset to their default values.
        """
        request = dto.SetServoTuningRequest(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
            paramset=paramset,
            tuning_params=tuning_params,
            set_unspecified_to_default=set_unspecified_to_default,
        )
        await call_async("servotuning/set_raw", request)

    def set_pid_tuning(
            self,
            paramset: ServoTuningParamset,
            p: float,
            i: float,
            d: float,
            fc: float
    ) -> PidTuning:
        """
        Sets the tuning of a paramset using the PID method.

        Args:
            paramset: The paramset to get tuning for.
            p: The proportional gain. Must be in units of N/m.
            i: The integral gain. Must be in units of N/m⋅s.
            d: The derivative gain. Must be in units of N⋅s/m.
            fc: The cutoff frequency. Must be in units of Hz.

        Returns:
            The PID representation of the current tuning after your changes have been applied.
        """
        request = dto.SetServoTuningPIDRequest(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
            paramset=paramset,
            p=p,
            i=i,
            d=d,
            fc=fc,
        )
        response = call(
            "servotuning/set_pid",
            request,
            PidTuning.from_binary)
        return response

    async def set_pid_tuning_async(
            self,
            paramset: ServoTuningParamset,
            p: float,
            i: float,
            d: float,
            fc: float
    ) -> PidTuning:
        """
        Sets the tuning of a paramset using the PID method.

        Args:
            paramset: The paramset to get tuning for.
            p: The proportional gain. Must be in units of N/m.
            i: The integral gain. Must be in units of N/m⋅s.
            d: The derivative gain. Must be in units of N⋅s/m.
            fc: The cutoff frequency. Must be in units of Hz.

        Returns:
            The PID representation of the current tuning after your changes have been applied.
        """
        request = dto.SetServoTuningPIDRequest(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
            paramset=paramset,
            p=p,
            i=i,
            d=d,
            fc=fc,
        )
        response = await call_async(
            "servotuning/set_pid",
            request,
            PidTuning.from_binary)
        return response

    def get_pid_tuning(
            self,
            paramset: ServoTuningParamset
    ) -> PidTuning:
        """
        Gets the PID representation of this paramset's servo tuning.

        Args:
            paramset: The paramset to get tuning for.

        Returns:
            The PID representation of the current tuning.
        """
        request = dto.ServoTuningRequest(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
            paramset=paramset,
        )
        response = call(
            "servotuning/get_pid",
            request,
            PidTuning.from_binary)
        return response

    async def get_pid_tuning_async(
            self,
            paramset: ServoTuningParamset
    ) -> PidTuning:
        """
        Gets the PID representation of this paramset's servo tuning.

        Args:
            paramset: The paramset to get tuning for.

        Returns:
            The PID representation of the current tuning.
        """
        request = dto.ServoTuningRequest(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
            paramset=paramset,
        )
        response = await call_async(
            "servotuning/get_pid",
            request,
            PidTuning.from_binary)
        return response

    def get_simple_tuning_param_definitions(
            self
    ) -> List[SimpleTuningParamDefinition]:
        """
        Gets the parameters that are required to tune this device.

        Returns:
            The tuning parameters.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
        )
        response = call(
            "servotuning/get_simple_params_definition",
            request,
            dto.GetSimpleTuningParamDefinitionResponse.from_binary)
        return response.params

    async def get_simple_tuning_param_definitions_async(
            self
    ) -> List[SimpleTuningParamDefinition]:
        """
        Gets the parameters that are required to tune this device.

        Returns:
            The tuning parameters.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
        )
        response = await call_async(
            "servotuning/get_simple_params_definition",
            request,
            dto.GetSimpleTuningParamDefinitionResponse.from_binary)
        return response.params

    def set_simple_tuning(
            self,
            paramset: ServoTuningParamset,
            tuning_params: List[ServoTuningParam],
            load_mass: float,
            carriage_mass: Optional[float] = None
    ) -> None:
        """
        Set the tuning of this device using the simple input method.

        Args:
            paramset: The paramset to set tuning for.
            tuning_params: The params used to tune this device.
                To get what parameters are expected, call GetSimpleTuningParamList.
                All values must be between 0 and 1.
            load_mass: The mass loaded on the stage (excluding the mass of the carriage itself) in kg.
            carriage_mass: The mass of the carriage in kg. If this value is not set the default carriage mass is used.
        """
        request = dto.SetSimpleTuning(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
            paramset=paramset,
            tuning_params=tuning_params,
            load_mass=load_mass,
            carriage_mass=carriage_mass,
        )
        call("servotuning/set_simple_tuning", request)

    async def set_simple_tuning_async(
            self,
            paramset: ServoTuningParamset,
            tuning_params: List[ServoTuningParam],
            load_mass: float,
            carriage_mass: Optional[float] = None
    ) -> None:
        """
        Set the tuning of this device using the simple input method.

        Args:
            paramset: The paramset to set tuning for.
            tuning_params: The params used to tune this device.
                To get what parameters are expected, call GetSimpleTuningParamList.
                All values must be between 0 and 1.
            load_mass: The mass loaded on the stage (excluding the mass of the carriage itself) in kg.
            carriage_mass: The mass of the carriage in kg. If this value is not set the default carriage mass is used.
        """
        request = dto.SetSimpleTuning(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
            paramset=paramset,
            tuning_params=tuning_params,
            load_mass=load_mass,
            carriage_mass=carriage_mass,
        )
        await call_async("servotuning/set_simple_tuning", request)

    def get_simple_tuning(
            self,
            paramset: ServoTuningParamset
    ) -> SimpleTuning:
        """
        Get the simple tuning parameters for this device.

        Args:
            paramset: The paramset to get tuning for.

        Returns:
            The simple tuning parameters.
        """
        request = dto.ServoTuningRequest(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
            paramset=paramset,
        )
        response = call(
            "servotuning/get_simple_tuning",
            request,
            SimpleTuning.from_binary)
        return response

    async def get_simple_tuning_async(
            self,
            paramset: ServoTuningParamset
    ) -> SimpleTuning:
        """
        Get the simple tuning parameters for this device.

        Args:
            paramset: The paramset to get tuning for.

        Returns:
            The simple tuning parameters.
        """
        request = dto.ServoTuningRequest(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
            paramset=paramset,
        )
        response = await call_async(
            "servotuning/get_simple_tuning",
            request,
            SimpleTuning.from_binary)
        return response

    def is_using_simple_tuning(
            self,
            paramset: ServoTuningParamset,
            tuning_params: List[ServoTuningParam],
            load_mass: float,
            carriage_mass: Optional[float] = None
    ) -> bool:
        """
        Deprecated: Use GetSimpleTuning instead.

        Checks if the provided simple tuning is being stored by this paramset.

        Args:
            paramset: The paramset to set tuning for.
            tuning_params: The params used to tune this device.
                To get what parameters are expected, call GetSimpleTuningParamList.
                All values must be between 0 and 1.
            load_mass: The mass loaded on the stage (excluding the mass of the carriage itself) in kg.
            carriage_mass: The mass of the carriage in kg. If this value is not set the default carriage mass is used.

        Returns:
            True if the provided simple tuning is currently stored in this paramset.
        """
        request = dto.SetSimpleTuning(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
            paramset=paramset,
            tuning_params=tuning_params,
            load_mass=load_mass,
            carriage_mass=carriage_mass,
        )
        response = call(
            "servotuning/is_using_simple_tuning",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    async def is_using_simple_tuning_async(
            self,
            paramset: ServoTuningParamset,
            tuning_params: List[ServoTuningParam],
            load_mass: float,
            carriage_mass: Optional[float] = None
    ) -> bool:
        """
        Deprecated: Use GetSimpleTuning instead.

        Checks if the provided simple tuning is being stored by this paramset.

        Args:
            paramset: The paramset to set tuning for.
            tuning_params: The params used to tune this device.
                To get what parameters are expected, call GetSimpleTuningParamList.
                All values must be between 0 and 1.
            load_mass: The mass loaded on the stage (excluding the mass of the carriage itself) in kg.
            carriage_mass: The mass of the carriage in kg. If this value is not set the default carriage mass is used.

        Returns:
            True if the provided simple tuning is currently stored in this paramset.
        """
        request = dto.SetSimpleTuning(
            interface_id=self.axis.device.connection.interface_id,
            device=self.axis.device.device_address,
            axis=self.axis.axis_number,
            paramset=paramset,
            tuning_params=tuning_params,
            load_mass=load_mass,
            carriage_mass=carriage_mass,
        )
        response = await call_async(
            "servotuning/is_using_simple_tuning",
            request,
            dto.BoolResponse.from_binary)
        return response.value



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long
from .alert_event_wrapper import AlertEventWrapper as AlertEventWrapper
from .autofocus_focus_request import AutofocusFocusRequest as AutofocusFocusRequest
from .autofocus_get_objective_params_request import AutofocusGetObjectiveParamsRequest as AutofocusGetObjectiveParamsRequest
from .autofocus_get_objective_params_response import AutofocusGetObjectiveParamsResponse as AutofocusGetObjectiveParamsResponse
from .autofocus_get_status_response import AutofocusGetStatusResponse as AutofocusGetStatusResponse
from .autofocus_set_objective_params_request import AutofocusSetObjectiveParamsRequest as AutofocusSetObjectiveParamsRequest
from .axes_empty_request import AxesEmptyRequest as AxesEmptyRequest
from .axes_get_setting_request import AxesGetSettingRequest as AxesGetSettingRequest
from .axes_move_request import AxesMoveRequest as AxesMoveRequest
from .axis_empty_request import AxisEmptyRequest as AxisEmptyRequest
from .axis_move_type import AxisMoveType as AxisMoveType
from .axis_to_string_request import AxisToStringRequest as AxisToStringRequest
from .binary_device_detect_request import BinaryDeviceDetectRequest as BinaryDeviceDetectRequest
from .binary_device_detect_response import BinaryDeviceDetectResponse as BinaryDeviceDetectResponse
from .binary_device_get_setting_request import BinaryDeviceGetSettingRequest as BinaryDeviceGetSettingRequest
from .binary_device_home_request import BinaryDeviceHomeRequest as BinaryDeviceHomeRequest
from .binary_device_move_request import BinaryDeviceMoveRequest as BinaryDeviceMoveRequest
from .binary_device_set_setting_request import BinaryDeviceSetSettingRequest as BinaryDeviceSetSettingRequest
from .binary_device_stop_request import BinaryDeviceStopRequest as BinaryDeviceStopRequest
from .binary_generic_with_units_request import BinaryGenericWithUnitsRequest as BinaryGenericWithUnitsRequest
from .binary_message_collection import BinaryMessageCollection as BinaryMessageCollection
from .binary_reply_only_event_wrapper import BinaryReplyOnlyEventWrapper as BinaryReplyOnlyEventWrapper
from .bool_response import BoolResponse as BoolResponse
from .can_set_state_axis_response import CanSetStateAxisResponse as CanSetStateAxisResponse
from .can_set_state_request import CanSetStateRequest as CanSetStateRequest
from .channel_on import ChannelOn as ChannelOn
from .channel_set_intensity import ChannelSetIntensity as ChannelSetIntensity
from .check_version_request import CheckVersionRequest as CheckVersionRequest
from .custom_interface_close_request import CustomInterfaceCloseRequest as CustomInterfaceCloseRequest
from .custom_interface_open_response import CustomInterfaceOpenResponse as CustomInterfaceOpenResponse
from .custom_interface_read_request import CustomInterfaceReadRequest as CustomInterfaceReadRequest
from .custom_interface_write_request import CustomInterfaceWriteRequest as CustomInterfaceWriteRequest
from .device_cancel_all_outputs_schedule_request import DeviceCancelAllOutputsScheduleRequest as DeviceCancelAllOutputsScheduleRequest
from .device_cancel_output_schedule_request import DeviceCancelOutputScheduleRequest as DeviceCancelOutputScheduleRequest
from .device_convert_setting_request import DeviceConvertSettingRequest as DeviceConvertSettingRequest
from .device_detect_request import DeviceDetectRequest as DeviceDetectRequest
from .device_detect_response import DeviceDetectResponse as DeviceDetectResponse
from .device_empty_request import DeviceEmptyRequest as DeviceEmptyRequest
from .device_get_all_analog_io_request import DeviceGetAllAnalogIORequest as DeviceGetAllAnalogIORequest
from .device_get_all_analog_io_response import DeviceGetAllAnalogIOResponse as DeviceGetAllAnalogIOResponse
from .device_get_all_digital_io_request import DeviceGetAllDigitalIORequest as DeviceGetAllDigitalIORequest
from .device_get_all_digital_io_response import DeviceGetAllDigitalIOResponse as DeviceGetAllDigitalIOResponse
from .device_get_analog_io_request import DeviceGetAnalogIORequest as DeviceGetAnalogIORequest
from .device_get_digital_io_request import DeviceGetDigitalIORequest as DeviceGetDigitalIORequest
from .device_get_setting_request import DeviceGetSettingRequest as DeviceGetSettingRequest
from .device_get_storage_request import DeviceGetStorageRequest as DeviceGetStorageRequest
from .device_get_warnings_request import DeviceGetWarningsRequest as DeviceGetWarningsRequest
from .device_get_warnings_response import DeviceGetWarningsResponse as DeviceGetWarningsResponse
from .device_home_request import DeviceHomeRequest as DeviceHomeRequest
from .device_identify_request import DeviceIdentifyRequest as DeviceIdentifyRequest
from .device_move_request import DeviceMoveRequest as DeviceMoveRequest
from .device_move_sin_request import DeviceMoveSinRequest as DeviceMoveSinRequest
from .device_multi_get_setting_request import DeviceMultiGetSettingRequest as DeviceMultiGetSettingRequest
from .device_on_all_request import DeviceOnAllRequest as DeviceOnAllRequest
from .device_on_all_response import DeviceOnAllResponse as DeviceOnAllResponse
from .device_restore_request import DeviceRestoreRequest as DeviceRestoreRequest
from .device_set_all_analog_outputs_request import DeviceSetAllAnalogOutputsRequest as DeviceSetAllAnalogOutputsRequest
from .device_set_all_analog_outputs_schedule_request import DeviceSetAllAnalogOutputsScheduleRequest as DeviceSetAllAnalogOutputsScheduleRequest
from .device_set_all_digital_outputs_request import DeviceSetAllDigitalOutputsRequest as DeviceSetAllDigitalOutputsRequest
from .device_set_all_digital_outputs_schedule_request import DeviceSetAllDigitalOutputsScheduleRequest as DeviceSetAllDigitalOutputsScheduleRequest
from .device_set_analog_output_request import DeviceSetAnalogOutputRequest as DeviceSetAnalogOutputRequest
from .device_set_analog_output_schedule_request import DeviceSetAnalogOutputScheduleRequest as DeviceSetAnalogOutputScheduleRequest
from .device_set_digital_output_request import DeviceSetDigitalOutputRequest as DeviceSetDigitalOutputRequest
from .device_set_digital_output_schedule_request import DeviceSetDigitalOutputScheduleRequest as DeviceSetDigitalOutputScheduleRequest
from .device_set_lowpass_filter_request import DeviceSetLowpassFilterRequest as DeviceSetLowpassFilterRequest
from .device_set_setting_request import DeviceSetSettingRequest as DeviceSetSettingRequest
from .device_set_setting_str_request import DeviceSetSettingStrRequest as DeviceSetSettingStrRequest
from .device_set_storage_bool_request import DeviceSetStorageBoolRequest as DeviceSetStorageBoolRequest
from .device_set_storage_number_request import DeviceSetStorageNumberRequest as DeviceSetStorageNumberRequest
from .device_set_storage_request import DeviceSetStorageRequest as DeviceSetStorageRequest
from .device_set_unit_conversions_request import DeviceSetUnitConversionsRequest as DeviceSetUnitConversionsRequest
from .device_stop_request import DeviceStopRequest as DeviceStopRequest
from .device_storage_list_keys_request import DeviceStorageListKeysRequest as DeviceStorageListKeysRequest
from .device_storage_request import DeviceStorageRequest as DeviceStorageRequest
from .device_type import DeviceType as DeviceType
from .device_wait_until_idle_request import DeviceWaitUntilIdleRequest as DeviceWaitUntilIdleRequest
from .disconnected_event import DisconnectedEvent as DisconnectedEvent
from .double_array_response import DoubleArrayResponse as DoubleArrayResponse
from .double_response import DoubleResponse as DoubleResponse
from .driver_enable_request import DriverEnableRequest as DriverEnableRequest
from .empty_autofocus_request import EmptyAutofocusRequest as EmptyAutofocusRequest
from .empty_request import EmptyRequest as EmptyRequest
from .errors import Errors as Errors
from .find_device_request import FindDeviceRequest as FindDeviceRequest
from .find_device_response import FindDeviceResponse as FindDeviceResponse
from .forget_devices_request import ForgetDevicesRequest as ForgetDevicesRequest
from .gateway_event import GatewayEvent as GatewayEvent
from .gateway_request import GatewayRequest as GatewayRequest
from .gateway_response import GatewayResponse as GatewayResponse
from .generic_binary_request import GenericBinaryRequest as GenericBinaryRequest
from .generic_command_request import GenericCommandRequest as GenericCommandRequest
from .generic_command_response_collection import GenericCommandResponseCollection as GenericCommandResponseCollection
from .get_all_io_port_labels_response import GetAllIoPortLabelsResponse as GetAllIoPortLabelsResponse
from .get_axis_setting_results import GetAxisSettingResults as GetAxisSettingResults
from .get_io_port_label_request import GetIoPortLabelRequest as GetIoPortLabelRequest
from .get_setting_results import GetSettingResults as GetSettingResults
from .get_simple_tuning_param_definition_response import GetSimpleTuningParamDefinitionResponse as GetSimpleTuningParamDefinitionResponse
from .int_array_response import IntArrayResponse as IntArrayResponse
from .interface_empty_request import InterfaceEmptyRequest as InterfaceEmptyRequest
from .interface_type import InterfaceType as InterfaceType
from .int_request import IntRequest as IntRequest
from .int_response import IntResponse as IntResponse
from .load_paramset import LoadParamset as LoadParamset
from .lockstep_disable_request import LockstepDisableRequest as LockstepDisableRequest
from .lockstep_empty_request import LockstepEmptyRequest as LockstepEmptyRequest
from .lockstep_enable_request import LockstepEnableRequest as LockstepEnableRequest
from .lockstep_get_axis_numbers_response import LockstepGetAxisNumbersResponse as LockstepGetAxisNumbersResponse
from .lockstep_get_request import LockstepGetRequest as LockstepGetRequest
from .lockstep_home_request import LockstepHomeRequest as LockstepHomeRequest
from .lockstep_move_request import LockstepMoveRequest as LockstepMoveRequest
from .lockstep_move_sin_request import LockstepMoveSinRequest as LockstepMoveSinRequest
from .lockstep_set_request import LockstepSetRequest as LockstepSetRequest
from .lockstep_stop_request import LockstepStopRequest as LockstepStopRequest
from .lockstep_wait_until_idle_request import LockstepWaitUntilIdleRequest as LockstepWaitUntilIdleRequest
from .microscope_config_response import MicroscopeConfigResponse as MicroscopeConfigResponse
from .microscope_empty_request import MicroscopeEmptyRequest as MicroscopeEmptyRequest
from .microscope_find_request import MicroscopeFindRequest as MicroscopeFindRequest
from .microscope_init_request import MicroscopeInitRequest as MicroscopeInitRequest
from .microscope_trigger_camera_request import MicroscopeTriggerCameraRequest as MicroscopeTriggerCameraRequest
from .objective_changer_change_request import ObjectiveChangerChangeRequest as ObjectiveChangerChangeRequest
from .objective_changer_create_response import ObjectiveChangerCreateResponse as ObjectiveChangerCreateResponse
from .objective_changer_get_current_response import ObjectiveChangerGetCurrentResponse as ObjectiveChangerGetCurrentResponse
from .objective_changer_request import ObjectiveChangerRequest as ObjectiveChangerRequest
from .objective_changer_set_request import ObjectiveChangerSetRequest as ObjectiveChangerSetRequest
from .open_binary_interface_request import OpenBinaryInterfaceRequest as OpenBinaryInterfaceRequest
from .open_interface_request import OpenInterfaceRequest as OpenInterfaceRequest
from .open_interface_response import OpenInterfaceResponse as OpenInterfaceResponse
from .oscilloscope_add_io_channel_request import OscilloscopeAddIoChannelRequest as OscilloscopeAddIoChannelRequest
from .oscilloscope_add_setting_channel_request import OscilloscopeAddSettingChannelRequest as OscilloscopeAddSettingChannelRequest
from .oscilloscope_data_get_request import OscilloscopeDataGetRequest as OscilloscopeDataGetRequest
from .oscilloscope_data_get_samples_response import OscilloscopeDataGetSamplesResponse as OscilloscopeDataGetSamplesResponse
from .oscilloscope_data_get_sample_time_request import OscilloscopeDataGetSampleTimeRequest as OscilloscopeDataGetSampleTimeRequest
from .oscilloscope_data_identifier import OscilloscopeDataIdentifier as OscilloscopeDataIdentifier
from .oscilloscope_read_response import OscilloscopeReadResponse as OscilloscopeReadResponse
from .oscilloscope_request import OscilloscopeRequest as OscilloscopeRequest
from .oscilloscope_start_request import OscilloscopeStartRequest as OscilloscopeStartRequest
from .prepare_command_request import PrepareCommandRequest as PrepareCommandRequest
from .process_on import ProcessOn as ProcessOn
from .pvt_point_request import PvtPointRequest as PvtPointRequest
from .renumber_request import RenumberRequest as RenumberRequest
from .response_type import ResponseType as ResponseType
from .servo_tuning_paramset_response import ServoTuningParamsetResponse as ServoTuningParamsetResponse
from .servo_tuning_request import ServoTuningRequest as ServoTuningRequest
from .set_device_db_source_request import SetDeviceDbSourceRequest as SetDeviceDbSourceRequest
from .set_interface_checksum_enabled_request import SetInterfaceChecksumEnabledRequest as SetInterfaceChecksumEnabledRequest
from .set_interface_timeout_request import SetInterfaceTimeoutRequest as SetInterfaceTimeoutRequest
from .set_internal_mode_request import SetInternalModeRequest as SetInternalModeRequest
from .set_io_port_label_request import SetIoPortLabelRequest as SetIoPortLabelRequest
from .set_log_output_request import SetLogOutputRequest as SetLogOutputRequest
from .set_process_controller_source import SetProcessControllerSource as SetProcessControllerSource
from .set_servo_tuning_pid_request import SetServoTuningPIDRequest as SetServoTuningPIDRequest
from .set_servo_tuning_request import SetServoTuningRequest as SetServoTuningRequest
from .set_simple_tuning import SetSimpleTuning as SetSimpleTuning
from .set_state_request import SetStateRequest as SetStateRequest
from .stream_arc_request import StreamArcRequest as StreamArcRequest
from .stream_buffer_erase_request import StreamBufferEraseRequest as StreamBufferEraseRequest
from .stream_buffer_get_content_request import StreamBufferGetContentRequest as StreamBufferGetContentRequest
from .stream_buffer_get_content_response import StreamBufferGetContentResponse as StreamBufferGetContentResponse
from .stream_buffer_list import StreamBufferList as StreamBufferList
from .stream_call_request import StreamCallRequest as StreamCallRequest
from .stream_cancel_all_outputs_schedule_request import StreamCancelAllOutputsScheduleRequest as StreamCancelAllOutputsScheduleRequest
from .stream_cancel_output_schedule_request import StreamCancelOutputScheduleRequest as StreamCancelOutputScheduleRequest
from .stream_circle_request import StreamCircleRequest as StreamCircleRequest
from .stream_empty_request import StreamEmptyRequest as StreamEmptyRequest
from .stream_generic_command_batch_request import StreamGenericCommandBatchRequest as StreamGenericCommandBatchRequest
from .stream_generic_command_request import StreamGenericCommandRequest as StreamGenericCommandRequest
from .stream_get_axes_response import StreamGetAxesResponse as StreamGetAxesResponse
from .stream_get_max_centripetal_acceleration_request import StreamGetMaxCentripetalAccelerationRequest as StreamGetMaxCentripetalAccelerationRequest
from .stream_get_max_speed_request import StreamGetMaxSpeedRequest as StreamGetMaxSpeedRequest
from .stream_get_max_tangential_acceleration_request import StreamGetMaxTangentialAccelerationRequest as StreamGetMaxTangentialAccelerationRequest
from .stream_line_request import StreamLineRequest as StreamLineRequest
from .stream_mode_response import StreamModeResponse as StreamModeResponse
from .stream_segment_type import StreamSegmentType as StreamSegmentType
from .stream_set_all_analog_outputs_request import StreamSetAllAnalogOutputsRequest as StreamSetAllAnalogOutputsRequest
from .stream_set_all_analog_outputs_schedule_request import StreamSetAllAnalogOutputsScheduleRequest as StreamSetAllAnalogOutputsScheduleRequest
from .stream_set_all_digital_outputs_request import StreamSetAllDigitalOutputsRequest as StreamSetAllDigitalOutputsRequest
from .stream_set_all_digital_outputs_schedule_request import StreamSetAllDigitalOutputsScheduleRequest as StreamSetAllDigitalOutputsScheduleRequest
from .stream_set_analog_output_request import StreamSetAnalogOutputRequest as StreamSetAnalogOutputRequest
from .stream_set_analog_output_schedule_request import StreamSetAnalogOutputScheduleRequest as StreamSetAnalogOutputScheduleRequest
from .stream_set_digital_output_request import StreamSetDigitalOutputRequest as StreamSetDigitalOutputRequest
from .stream_set_digital_output_schedule_request import StreamSetDigitalOutputScheduleRequest as StreamSetDigitalOutputScheduleRequest
from .stream_set_hold_request import StreamSetHoldRequest as StreamSetHoldRequest
from .stream_set_max_centripetal_acceleration_request import StreamSetMaxCentripetalAccelerationRequest as StreamSetMaxCentripetalAccelerationRequest
from .stream_set_max_speed_request import StreamSetMaxSpeedRequest as StreamSetMaxSpeedRequest
from .stream_set_max_tangential_acceleration_request import StreamSetMaxTangentialAccelerationRequest as StreamSetMaxTangentialAccelerationRequest
from .stream_setup_live_composite_request import StreamSetupLiveCompositeRequest as StreamSetupLiveCompositeRequest
from .stream_setup_live_request import StreamSetupLiveRequest as StreamSetupLiveRequest
from .stream_setup_store_arbitrary_axes_request import StreamSetupStoreArbitraryAxesRequest as StreamSetupStoreArbitraryAxesRequest
from .stream_setup_store_composite_request import StreamSetupStoreCompositeRequest as StreamSetupStoreCompositeRequest
from .stream_setup_store_request import StreamSetupStoreRequest as StreamSetupStoreRequest
from .stream_wait_analog_input_request import StreamWaitAnalogInputRequest as StreamWaitAnalogInputRequest
from .stream_wait_digital_input_request import StreamWaitDigitalInputRequest as StreamWaitDigitalInputRequest
from .stream_wait_request import StreamWaitRequest as StreamWaitRequest
from .stream_wait_until_idle_request import StreamWaitUntilIdleRequest as StreamWaitUntilIdleRequest
from .string_array_response import StringArrayResponse as StringArrayResponse
from .string_response import StringResponse as StringResponse
from .test_event import TestEvent as TestEvent
from .test_request import TestRequest as TestRequest
from .test_response import TestResponse as TestResponse
from .test_response_long import TestResponseLong as TestResponseLong
from .toggle_device_db_store_request import ToggleDeviceDbStoreRequest as ToggleDeviceDbStoreRequest
from .tools_list_serial_ports_response import ToolsListSerialPortsResponse as ToolsListSerialPortsResponse
from .translator_create_from_device_request import TranslatorCreateFromDeviceRequest as TranslatorCreateFromDeviceRequest
from .translator_create_live_request import TranslatorCreateLiveRequest as TranslatorCreateLiveRequest
from .translator_create_request import TranslatorCreateRequest as TranslatorCreateRequest
from .translator_create_response import TranslatorCreateResponse as TranslatorCreateResponse
from .translator_empty_request import TranslatorEmptyRequest as TranslatorEmptyRequest
from .translator_flush_live_request import TranslatorFlushLiveRequest as TranslatorFlushLiveRequest
from .translator_flush_response import TranslatorFlushResponse as TranslatorFlushResponse
from .translator_get_axis_offset_request import TranslatorGetAxisOffsetRequest as TranslatorGetAxisOffsetRequest
from .translator_get_axis_position_request import TranslatorGetAxisPositionRequest as TranslatorGetAxisPositionRequest
from .translator_set_axis_position_request import TranslatorSetAxisPositionRequest as TranslatorSetAxisPositionRequest
from .translator_set_feed_rate_override_request import TranslatorSetFeedRateOverrideRequest as TranslatorSetFeedRateOverrideRequest
from .translator_set_traverse_rate_request import TranslatorSetTraverseRateRequest as TranslatorSetTraverseRateRequest
from .translator_translate_request import TranslatorTranslateRequest as TranslatorTranslateRequest
from .trigger_clear_action_request import TriggerClearActionRequest as TriggerClearActionRequest
from .trigger_empty_request import TriggerEmptyRequest as TriggerEmptyRequest
from .trigger_enabled_states import TriggerEnabledStates as TriggerEnabledStates
from .trigger_enable_request import TriggerEnableRequest as TriggerEnableRequest
from .trigger_fire_at_interval_request import TriggerFireAtIntervalRequest as TriggerFireAtIntervalRequest
from .trigger_fire_when_distance_travelled_request import TriggerFireWhenDistanceTravelledRequest as TriggerFireWhenDistanceTravelledRequest
from .trigger_fire_when_io_request import TriggerFireWhenIoRequest as TriggerFireWhenIoRequest
from .trigger_fire_when_request import TriggerFireWhenRequest as TriggerFireWhenRequest
from .trigger_fire_when_setting_request import TriggerFireWhenSettingRequest as TriggerFireWhenSettingRequest
from .trigger_on_fire_request import TriggerOnFireRequest as TriggerOnFireRequest
from .trigger_on_fire_set_request import TriggerOnFireSetRequest as TriggerOnFireSetRequest
from .trigger_on_fire_set_to_setting_request import TriggerOnFireSetToSettingRequest as TriggerOnFireSetToSettingRequest
from .trigger_set_label_request import TriggerSetLabelRequest as TriggerSetLabelRequest
from .trigger_states import TriggerStates as TriggerStates
from .unit_get_enum_request import UnitGetEnumRequest as UnitGetEnumRequest
from .unit_get_enum_response import UnitGetEnumResponse as UnitGetEnumResponse
from .unit_get_symbol_request import UnitGetSymbolRequest as UnitGetSymbolRequest
from .unit_get_symbol_response import UnitGetSymbolResponse as UnitGetSymbolResponse
from .unknown_binary_response_event_wrapper import UnknownBinaryResponseEventWrapper as UnknownBinaryResponseEventWrapper
from .unknown_response_event_wrapper import UnknownResponseEventWrapper as UnknownResponseEventWrapper
from .wait_to_clear_warnings_request import WaitToClearWarningsRequest as WaitToClearWarningsRequest
from .wait_to_respond_request import WaitToRespondRequest as WaitToRespondRequest
from .wdi_generic_request import WdiGenericRequest as WdiGenericRequest
from .wdi_get_status_response import WdiGetStatusResponse as WdiGetStatusResponse



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, List
import decimal
from collections.abc import Iterable
import zaber_bson
from .axis_definition import AxisDefinition
from ..measurement import Measurement


@dataclass
class DeviceDefinition:
    """
    Holds information about device and its axes for purpose of a translator.
    """

    device_id: int
    """
    Device ID of the controller.
    Can be obtained from device settings.
    """

    axes: List[AxisDefinition]
    """
    Applicable axes of the device.
    """

    max_speed: Measurement
    """
    The smallest of each axis' maxspeed setting value.
    This value becomes the traverse rate of the translator.
    """

    @staticmethod
    def zero_values() -> 'DeviceDefinition':
        return DeviceDefinition(
            device_id=0,
            axes=[],
            max_speed=Measurement.zero_values(),
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'DeviceDefinition':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return DeviceDefinition.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'deviceId': int(self.device_id),
            'axes': [item.to_dict() for item in self.axes] if self.axes is not None else [],
            'maxSpeed': self.max_speed.to_dict(),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'DeviceDefinition':
        return DeviceDefinition(
            device_id=data.get('deviceId'),  # type: ignore
            axes=[AxisDefinition.from_dict(item) for item in data.get('axes')],  # type: ignore
            max_speed=Measurement.from_dict(data.get('maxSpeed')),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.device_id is None:
            raise ValueError(f'Property "DeviceId" of "DeviceDefinition" is None.')

        if not isinstance(self.device_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "DeviceId" of "DeviceDefinition" is not a number.')

        if int(self.device_id) != self.device_id:
            raise ValueError(f'Property "DeviceId" of "DeviceDefinition" is not integer value.')

        if self.axes is not None:
            if not isinstance(self.axes, Iterable):
                raise ValueError('Property "Axes" of "DeviceDefinition" is not iterable.')

            for i, axes_item in enumerate(self.axes):
                if axes_item is None:
                    raise ValueError(f'Item {i} in property "Axes" of "DeviceDefinition" is None.')

                if not isinstance(axes_item, AxisDefinition):
                    raise ValueError(f'Item {i} in property "Axes" of "DeviceDefinition" is not an instance of "AxisDefinition".')

                axes_item.validate()

        if self.max_speed is None:
            raise ValueError(f'Property "MaxSpeed" of "DeviceDefinition" is None.')

        if not isinstance(self.max_speed, Measurement):
            raise ValueError(f'Property "MaxSpeed" of "DeviceDefinition" is not an instance of "Measurement".')

        self.max_speed.validate()



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, Optional
import zaber_bson
from ..log_output_mode import LogOutputMode


@dataclass
class SetLogOutputRequest:

    mode: LogOutputMode = next(first for first in LogOutputMode)

    file_path: Optional[str] = None

    @staticmethod
    def zero_values() -> 'SetLogOutputRequest':
        return SetLogOutputRequest(
            mode=next(first for first in LogOutputMode),
            file_path=None,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'SetLogOutputRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return SetLogOutputRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'mode': self.mode.value,
            'filePath': str(self.file_path) if self.file_path is not None else None,
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'SetLogOutputRequest':
        return SetLogOutputRequest(
            mode=LogOutputMode(data.get('mode')),  # type: ignore
            file_path=data.get('filePath'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.mode is None:
            raise ValueError(f'Property "Mode" of "SetLogOutputRequest" is None.')

        if not isinstance(self.mode, LogOutputMode):
            raise ValueError(f'Property "Mode" of "SetLogOutputRequest" is not an instance of "LogOutputMode".')

        if self.file_path is not None:
            if not isinstance(self.file_path, str):
                raise ValueError(f'Property "FilePath" of "SetLogOutputRequest" is not a string.')



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #
from typing import TYPE_CHECKING, List
from ..dto import requests as dto
from ..call import call, call_async

if TYPE_CHECKING:
    from .device import Device


class PvtBuffer:
    """
    Represents a PVT buffer with this number on a device.
    A PVT buffer is a place to store a queue of PVT actions.
    """

    @property
    def device(self) -> 'Device':
        """
        The Device this buffer exists on.
        """
        return self._device

    @property
    def buffer_id(self) -> int:
        """
        The number identifying the buffer on the device.
        """
        return self._buffer_id

    def __init__(self, device: 'Device', buffer_id: int):
        self._device = device
        self._buffer_id = buffer_id

    def get_content(
            self
    ) -> List[str]:
        """
        Gets the buffer contents as an array of strings.

        Returns:
            A string array containing all the PVT commands stored in the buffer.
        """
        request = dto.StreamBufferGetContentRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            buffer_id=self.buffer_id,
            pvt=True,
        )
        response = call(
            "device/stream_buffer_get_content",
            request,
            dto.StreamBufferGetContentResponse.from_binary)
        return response.buffer_lines

    async def get_content_async(
            self
    ) -> List[str]:
        """
        Gets the buffer contents as an array of strings.

        Returns:
            A string array containing all the PVT commands stored in the buffer.
        """
        request = dto.StreamBufferGetContentRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            buffer_id=self.buffer_id,
            pvt=True,
        )
        response = await call_async(
            "device/stream_buffer_get_content",
            request,
            dto.StreamBufferGetContentResponse.from_binary)
        return response.buffer_lines

    def erase(
            self
    ) -> None:
        """
        Erases the contents of the buffer.
        This method fails if there is a PVT sequence writing to the buffer.
        """
        request = dto.StreamBufferEraseRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            buffer_id=self.buffer_id,
            pvt=True,
        )
        call("device/stream_buffer_erase", request)

    async def erase_async(
            self
    ) -> None:
        """
        Erases the contents of the buffer.
        This method fails if there is a PVT sequence writing to the buffer.
        """
        request = dto.StreamBufferEraseRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            buffer_id=self.buffer_id,
            pvt=True,
        )
        await call_async("device/stream_buffer_erase", request)



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #
from typing import TYPE_CHECKING, List
from ..dto import requests as dto
from ..call import call, call_async
from .trigger import Trigger
from ..dto.ascii.trigger_state import TriggerState
from ..dto.ascii.trigger_enabled_state import TriggerEnabledState

if TYPE_CHECKING:
    from .device import Device


class Triggers:
    """
    Class providing access to device triggers.
    Please note that the Triggers API is currently an experimental feature.
    Requires at least Firmware 7.06.
    """

    @property
    def device(self) -> 'Device':
        """
        Device that these triggers belong to.
        """
        return self._device

    def __init__(self, device: 'Device'):
        self._device = device

    def get_number_of_triggers(
            self
    ) -> int:
        """
        Get the number of triggers for this device.

        Returns:
            Number of triggers for this device.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="trigger.numtriggers",
        )
        response = call(
            "triggers/get_setting",
            request,
            dto.IntResponse.from_binary)
        return response.value

    async def get_number_of_triggers_async(
            self
    ) -> int:
        """
        Get the number of triggers for this device.

        Returns:
            Number of triggers for this device.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="trigger.numtriggers",
        )
        response = await call_async(
            "triggers/get_setting",
            request,
            dto.IntResponse.from_binary)
        return response.value

    def get_number_of_actions(
            self
    ) -> int:
        """
        Get the number of actions for each trigger for this device.

        Returns:
            Number of actions for each trigger for this device.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="trigger.numactions",
        )
        response = call(
            "triggers/get_setting",
            request,
            dto.IntResponse.from_binary)
        return response.value

    async def get_number_of_actions_async(
            self
    ) -> int:
        """
        Get the number of actions for each trigger for this device.

        Returns:
            Number of actions for each trigger for this device.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
            setting="trigger.numactions",
        )
        response = await call_async(
            "triggers/get_setting",
            request,
            dto.IntResponse.from_binary)
        return response.value

    def get_trigger(
            self,
            trigger_number: int
    ) -> 'Trigger':
        """
        Get a specific trigger for this device.

        Args:
            trigger_number: The number of the trigger to control. Trigger numbers start at 1.

        Returns:
            Trigger instance.
        """
        if trigger_number <= 0:
            raise ValueError('Invalid value; triggers are numbered from 1.')

        return Trigger(self.device, trigger_number)

    def get_trigger_states(
            self
    ) -> List[TriggerState]:
        """
        Get the state for every trigger for this device.

        Returns:
            Complete state for every trigger.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
        )
        response = call(
            "triggers/get_trigger_states",
            request,
            dto.TriggerStates.from_binary)
        return response.states

    async def get_trigger_states_async(
            self
    ) -> List[TriggerState]:
        """
        Get the state for every trigger for this device.

        Returns:
            Complete state for every trigger.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
        )
        response = await call_async(
            "triggers/get_trigger_states",
            request,
            dto.TriggerStates.from_binary)
        return response.states

    def get_enabled_states(
            self
    ) -> List[TriggerEnabledState]:
        """
        Gets the enabled state for every trigger for this device.

        Returns:
            Whether triggers are enabled and the number of times they will fire.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
        )
        response = call(
            "triggers/get_enabled_states",
            request,
            dto.TriggerEnabledStates.from_binary)
        return response.states

    async def get_enabled_states_async(
            self
    ) -> List[TriggerEnabledState]:
        """
        Gets the enabled state for every trigger for this device.

        Returns:
            Whether triggers are enabled and the number of times they will fire.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
        )
        response = await call_async(
            "triggers/get_enabled_states",
            request,
            dto.TriggerEnabledStates.from_binary)
        return response.states

    def get_all_labels(
            self
    ) -> List[str]:
        """
        Gets the labels for every trigger for this device.

        Returns:
            The labels for every trigger for this device. If a trigger has no label, the value will be an empty string.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
        )
        response = call(
            "triggers/get_all_labels",
            request,
            dto.StringArrayResponse.from_binary)
        return response.values

    async def get_all_labels_async(
            self
    ) -> List[str]:
        """
        Gets the labels for every trigger for this device.

        Returns:
            The labels for every trigger for this device. If a trigger has no label, the value will be an empty string.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
        )
        response = await call_async(
            "triggers/get_all_labels",
            request,
            dto.StringArrayResponse.from_binary)
        return response.values



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, Optional
import decimal
import zaber_bson
from ..measurement import Measurement


@dataclass
class AxisTransformation:
    """
    Represents a transformation of a translator axis.
    """

    axis_letter: str
    """
    Letter of the translator axis (X,Y,Z,A,B,C,E).
    """

    scaling: Optional[float] = None
    """
    Scaling factor.
    """

    translation: Optional[Measurement] = None
    """
    Translation distance.
    """

    @staticmethod
    def zero_values() -> 'AxisTransformation':
        return AxisTransformation(
            axis_letter="",
            scaling=None,
            translation=None,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'AxisTransformation':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return AxisTransformation.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'axisLetter': str(self.axis_letter or ''),
            'scaling': float(self.scaling) if self.scaling is not None else None,
            'translation': self.translation.to_dict() if self.translation is not None else None,
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'AxisTransformation':
        return AxisTransformation(
            axis_letter=data.get('axisLetter'),  # type: ignore
            scaling=data.get('scaling'),  # type: ignore
            translation=Measurement.from_dict(data.get('translation')) if data.get('translation') is not None else None,  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.axis_letter is not None:
            if not isinstance(self.axis_letter, str):
                raise ValueError(f'Property "AxisLetter" of "AxisTransformation" is not a string.')

        if self.scaling is not None:
            if not isinstance(self.scaling, (int, float, decimal.Decimal)):
                raise ValueError(f'Property "Scaling" of "AxisTransformation" is not a number.')

        if self.translation is not None:
            if not isinstance(self.translation, Measurement):
                raise ValueError(f'Property "Translation" of "AxisTransformation" is not an instance of "Measurement".')

            self.translation.validate()



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, List
import decimal
from collections.abc import Iterable
import zaber_bson


@dataclass
class SetStateAxisResponse:
    """
    An object containing any non-blocking issues encountered when loading a saved state to an axis.
    """

    warnings: List[str]
    """
    The warnings encountered when applying this state to the given axis.
    """

    axis_number: int
    """
    The number of the axis that was set.
    """

    @staticmethod
    def zero_values() -> 'SetStateAxisResponse':
        return SetStateAxisResponse(
            warnings=[],
            axis_number=0,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'SetStateAxisResponse':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return SetStateAxisResponse.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'warnings': [str(item or '') for item in self.warnings] if self.warnings is not None else [],
            'axisNumber': int(self.axis_number),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'SetStateAxisResponse':
        return SetStateAxisResponse(
            warnings=data.get('warnings'),  # type: ignore
            axis_number=data.get('axisNumber'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.warnings is not None:
            if not isinstance(self.warnings, Iterable):
                raise ValueError('Property "Warnings" of "SetStateAxisResponse" is not iterable.')

            for i, warnings_item in enumerate(self.warnings):
                if warnings_item is not None:
                    if not isinstance(warnings_item, str):
                        raise ValueError(f'Item {i} in property "Warnings" of "SetStateAxisResponse" is not a string.')

        if self.axis_number is None:
            raise ValueError(f'Property "AxisNumber" of "SetStateAxisResponse" is None.')

        if not isinstance(self.axis_number, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "AxisNumber" of "SetStateAxisResponse" is not a number.')

        if int(self.axis_number) != self.axis_number:
            raise ValueError(f'Property "AxisNumber" of "SetStateAxisResponse" is not integer value.')



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass, field
from typing import Any, Dict, List
import decimal
from collections.abc import Iterable
import zaber_bson
from ..measurement import Measurement


@dataclass
class PrepareCommandRequest:

    interface_id: int = 0

    device: int = 0

    axis: int = 0

    command_template: str = ""

    parameters: List[Measurement] = field(default_factory=list)

    @staticmethod
    def zero_values() -> 'PrepareCommandRequest':
        return PrepareCommandRequest(
            interface_id=0,
            device=0,
            axis=0,
            command_template="",
            parameters=[],
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'PrepareCommandRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return PrepareCommandRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'interfaceId': int(self.interface_id),
            'device': int(self.device),
            'axis': int(self.axis),
            'commandTemplate': str(self.command_template or ''),
            'parameters': [item.to_dict() for item in self.parameters] if self.parameters is not None else [],
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'PrepareCommandRequest':
        return PrepareCommandRequest(
            interface_id=data.get('interfaceId'),  # type: ignore
            device=data.get('device'),  # type: ignore
            axis=data.get('axis'),  # type: ignore
            command_template=data.get('commandTemplate'),  # type: ignore
            parameters=[Measurement.from_dict(item) for item in data.get('parameters')],  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.interface_id is None:
            raise ValueError(f'Property "InterfaceId" of "PrepareCommandRequest" is None.')

        if not isinstance(self.interface_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "InterfaceId" of "PrepareCommandRequest" is not a number.')

        if int(self.interface_id) != self.interface_id:
            raise ValueError(f'Property "InterfaceId" of "PrepareCommandRequest" is not integer value.')

        if self.device is None:
            raise ValueError(f'Property "Device" of "PrepareCommandRequest" is None.')

        if not isinstance(self.device, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Device" of "PrepareCommandRequest" is not a number.')

        if int(self.device) != self.device:
            raise ValueError(f'Property "Device" of "PrepareCommandRequest" is not integer value.')

        if self.axis is None:
            raise ValueError(f'Property "Axis" of "PrepareCommandRequest" is None.')

        if not isinstance(self.axis, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Axis" of "PrepareCommandRequest" is not a number.')

        if int(self.axis) != self.axis:
            raise ValueError(f'Property "Axis" of "PrepareCommandRequest" is not integer value.')

        if self.command_template is not None:
            if not isinstance(self.command_template, str):
                raise ValueError(f'Property "CommandTemplate" of "PrepareCommandRequest" is not a string.')

        if self.parameters is not None:
            if not isinstance(self.parameters, Iterable):
                raise ValueError('Property "Parameters" of "PrepareCommandRequest" is not iterable.')

            for i, parameters_item in enumerate(self.parameters):
                if parameters_item is None:
                    raise ValueError(f'Item {i} in property "Parameters" of "PrepareCommandRequest" is None.')

                if not isinstance(parameters_item, Measurement):
                    raise ValueError(f'Item {i} in property "Parameters" of "PrepareCommandRequest" is not an instance of "Measurement".')

                parameters_item.validate()



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass, field
from typing import Any, Dict, Optional
import zaber_bson
from ..gcode.device_definition import DeviceDefinition
from ..gcode.translator_config import TranslatorConfig


@dataclass
class TranslatorCreateRequest:

    definition: DeviceDefinition = field(default_factory=DeviceDefinition.zero_values)

    config: Optional[TranslatorConfig] = None

    @staticmethod
    def zero_values() -> 'TranslatorCreateRequest':
        return TranslatorCreateRequest(
            definition=DeviceDefinition.zero_values(),
            config=None,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'TranslatorCreateRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return TranslatorCreateRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'definition': self.definition.to_dict(),
            'config': self.config.to_dict() if self.config is not None else None,
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'TranslatorCreateRequest':
        return TranslatorCreateRequest(
            definition=DeviceDefinition.from_dict(data.get('definition')),  # type: ignore
            config=TranslatorConfig.from_dict(data.get('config')) if data.get('config') is not None else None,  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.definition is None:
            raise ValueError(f'Property "Definition" of "TranslatorCreateRequest" is None.')

        if not isinstance(self.definition, DeviceDefinition):
            raise ValueError(f'Property "Definition" of "TranslatorCreateRequest" is not an instance of "DeviceDefinition".')

        self.definition.validate()

        if self.config is not None:
            if not isinstance(self.config, TranslatorConfig):
                raise ValueError(f'Property "Config" of "TranslatorCreateRequest" is not an instance of "TranslatorConfig".')

            self.config.validate()



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson
from ...units import Units, UnitsAndLiterals, units_from_literals


@dataclass
class TranslatorGetAxisOffsetRequest:

    translator_id: int = 0

    coordinate_system: str = ""

    axis: str = ""

    unit: UnitsAndLiterals = Units.NATIVE

    @staticmethod
    def zero_values() -> 'TranslatorGetAxisOffsetRequest':
        return TranslatorGetAxisOffsetRequest(
            translator_id=0,
            coordinate_system="",
            axis="",
            unit=Units.NATIVE,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'TranslatorGetAxisOffsetRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return TranslatorGetAxisOffsetRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'translatorId': int(self.translator_id),
            'coordinateSystem': str(self.coordinate_system or ''),
            'axis': str(self.axis or ''),
            'unit': units_from_literals(self.unit).value,
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'TranslatorGetAxisOffsetRequest':
        return TranslatorGetAxisOffsetRequest(
            translator_id=data.get('translatorId'),  # type: ignore
            coordinate_system=data.get('coordinateSystem'),  # type: ignore
            axis=data.get('axis'),  # type: ignore
            unit=Units(data.get('unit')),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.translator_id is None:
            raise ValueError(f'Property "TranslatorId" of "TranslatorGetAxisOffsetRequest" is None.')

        if not isinstance(self.translator_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "TranslatorId" of "TranslatorGetAxisOffsetRequest" is not a number.')

        if int(self.translator_id) != self.translator_id:
            raise ValueError(f'Property "TranslatorId" of "TranslatorGetAxisOffsetRequest" is not integer value.')

        if self.coordinate_system is not None:
            if not isinstance(self.coordinate_system, str):
                raise ValueError(f'Property "CoordinateSystem" of "TranslatorGetAxisOffsetRequest" is not a string.')

        if self.axis is not None:
            if not isinstance(self.axis, str):
                raise ValueError(f'Property "Axis" of "TranslatorGetAxisOffsetRequest" is not a string.')

        if self.unit is None:
            raise ValueError(f'Property "Unit" of "TranslatorGetAxisOffsetRequest" is None.')

        if not isinstance(self.unit, (Units, str)):
            raise ValueError(f'Property "Unit" of "TranslatorGetAxisOffsetRequest" is not Units.')



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from .motion_lib_exception import MotionLibException


class InvalidRequestDataException(MotionLibException):
    """
    Used for internal error handling.
    Indicates passing values of incorrect type from scripting languages or mixed library binary files.
    """



class MotionLibException(Exception):

    @property
    def message(self) -> str:
        """
        Error message of the exception.
        """
        return self._message

    def __init__(self, message: str):
        self._message = message
        Exception.__init__(self, self._message)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}: {self._message}"



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import Optional
from ..call import call, call_async, call_sync
from ..dto import requests as dto
from ..dto.measurement import Measurement
from ..ascii import Connection, Device, Axis
from ..units import Units, LengthUnits


class ObjectiveChanger:
    """
    Represents an objective changer of a microscope.
    Unstable. Expect breaking changes in future releases.
    Requires at least Firmware 7.32.
    """

    @property
    def turret(self) -> Device:
        """
        Device address of the turret.
        """
        return self._turret

    @property
    def focus_axis(self) -> Axis:
        """
        The focus axis.
        """
        return self._focus_axis

    def __init__(self, turret: Device, focus_axis: Axis):
        """
        Creates instance of `ObjectiveChanger` based on the given device.
        If the device is identified, this constructor will ensure it is an objective changer.
        """
        self._turret = turret
        self._focus_axis = focus_axis
        self.__verify_is_changer()

    @staticmethod
    def find(
            connection: Connection,
            turret_address: int = 0,
            focus_address: int = 0
    ) -> 'ObjectiveChanger':
        """
        Deprecated: Use microscope's `Find` method instead or instantiate manually.

        Finds an objective changer on a connection.
        In case of conflict, specify the optional device addresses.
        Devices on the connection must be identified.

        Args:
            connection: Connection on which to detect the objective changer.
            turret_address: Optional device address of the turret device (X-MOR).
            focus_address: Optional device address of the focus device (X-LDA).

        Returns:
            New instance of objective changer.
        """
        request = dto.ObjectiveChangerRequest(
            interface_id=connection.interface_id,
            turret_address=turret_address,
            focus_address=focus_address,
        )
        response = call(
            "objective_changer/detect",
            request,
            dto.ObjectiveChangerCreateResponse.from_binary)
        return ObjectiveChanger(
            Device(connection, response.turret),
            Axis(Device(connection, response.focus_address), response.focus_axis))

    @staticmethod
    async def find_async(
            connection: Connection,
            turret_address: int = 0,
            focus_address: int = 0
    ) -> 'ObjectiveChanger':
        """
        Deprecated: Use microscope's `Find` method instead or instantiate manually.

        Finds an objective changer on a connection.
        In case of conflict, specify the optional device addresses.
        Devices on the connection must be identified.

        Args:
            connection: Connection on which to detect the objective changer.
            turret_address: Optional device address of the turret device (X-MOR).
            focus_address: Optional device address of the focus device (X-LDA).

        Returns:
            New instance of objective changer.
        """
        request = dto.ObjectiveChangerRequest(
            interface_id=connection.interface_id,
            turret_address=turret_address,
            focus_address=focus_address,
        )
        response = await call_async(
            "objective_changer/detect",
            request,
            dto.ObjectiveChangerCreateResponse.from_binary)
        return ObjectiveChanger(
            Device(connection, response.turret),
            Axis(Device(connection, response.focus_address), response.focus_axis))

    def change(
            self,
            objective: int,
            focus_offset: Optional[Measurement] = None
    ) -> None:
        """
        Changes the objective.
        Runs a sequence of movements switching from the current objective to the new one.
        The focus stage moves to the focus datum after the objective change.

        Args:
            objective: Objective number starting from 1.
            focus_offset: Optional offset from the focus datum.
        """
        request = dto.ObjectiveChangerChangeRequest(
            interface_id=self.turret.connection.interface_id,
            turret_address=self.turret.device_address,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            objective=objective,
            focus_offset=focus_offset,
        )
        call("objective_changer/change", request)

    async def change_async(
            self,
            objective: int,
            focus_offset: Optional[Measurement] = None
    ) -> None:
        """
        Changes the objective.
        Runs a sequence of movements switching from the current objective to the new one.
        The focus stage moves to the focus datum after the objective change.

        Args:
            objective: Objective number starting from 1.
            focus_offset: Optional offset from the focus datum.
        """
        request = dto.ObjectiveChangerChangeRequest(
            interface_id=self.turret.connection.interface_id,
            turret_address=self.turret.device_address,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            objective=objective,
            focus_offset=focus_offset,
        )
        await call_async("objective_changer/change", request)

    def release(
            self
    ) -> None:
        """
        Moves the focus stage out of the turret releasing the current objective.
        """
        request = dto.ObjectiveChangerRequest(
            interface_id=self.turret.connection.interface_id,
            turret_address=self.turret.device_address,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
        )
        call("objective_changer/release", request)

    async def release_async(
            self
    ) -> None:
        """
        Moves the focus stage out of the turret releasing the current objective.
        """
        request = dto.ObjectiveChangerRequest(
            interface_id=self.turret.connection.interface_id,
            turret_address=self.turret.device_address,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
        )
        await call_async("objective_changer/release", request)

    def get_current_objective(
            self
    ) -> int:
        """
        Returns current objective number starting from 1.
        The value of 0 indicates that the position is either unknown or between two objectives.

        Returns:
            Current objective number starting from 1 or 0 if not applicable.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.turret.connection.interface_id,
            device=self.turret.device_address,
            axis=1,
        )
        response = call(
            "device/get_index_position",
            request,
            dto.IntResponse.from_binary)
        return response.value

    async def get_current_objective_async(
            self
    ) -> int:
        """
        Returns current objective number starting from 1.
        The value of 0 indicates that the position is either unknown or between two objectives.

        Returns:
            Current objective number starting from 1 or 0 if not applicable.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.turret.connection.interface_id,
            device=self.turret.device_address,
            axis=1,
        )
        response = await call_async(
            "device/get_index_position",
            request,
            dto.IntResponse.from_binary)
        return response.value

    def get_number_of_objectives(
            self
    ) -> int:
        """
        Gets number of objectives that the turret can accommodate.

        Returns:
            Number of positions.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.turret.connection.interface_id,
            device=self.turret.device_address,
            axis=1,
        )
        response = call(
            "device/get_index_count",
            request,
            dto.IntResponse.from_binary)
        return response.value

    async def get_number_of_objectives_async(
            self
    ) -> int:
        """
        Gets number of objectives that the turret can accommodate.

        Returns:
            Number of positions.
        """
        request = dto.AxisEmptyRequest(
            interface_id=self.turret.connection.interface_id,
            device=self.turret.device_address,
            axis=1,
        )
        response = await call_async(
            "device/get_index_count",
            request,
            dto.IntResponse.from_binary)
        return response.value

    def get_focus_datum(
            self,
            unit: LengthUnits = Units.NATIVE
    ) -> float:
        """
        Gets the focus datum.
        The focus datum is the position that the focus stage moves to after an objective change.
        It is backed by the limit.home.offset setting.

        Args:
            unit: Units of datum.

        Returns:
            The datum.
        """
        request = dto.ObjectiveChangerSetRequest(
            interface_id=self.turret.connection.interface_id,
            turret_address=self.turret.device_address,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            unit=unit,
        )
        response = call(
            "objective_changer/get_datum",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    async def get_focus_datum_async(
            self,
            unit: LengthUnits = Units.NATIVE
    ) -> float:
        """
        Gets the focus datum.
        The focus datum is the position that the focus stage moves to after an objective change.
        It is backed by the limit.home.offset setting.

        Args:
            unit: Units of datum.

        Returns:
            The datum.
        """
        request = dto.ObjectiveChangerSetRequest(
            interface_id=self.turret.connection.interface_id,
            turret_address=self.turret.device_address,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            unit=unit,
        )
        response = await call_async(
            "objective_changer/get_datum",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def set_focus_datum(
            self,
            datum: float,
            unit: LengthUnits = Units.NATIVE
    ) -> None:
        """
        Sets the focus datum.
        The focus datum is the position that the focus stage moves to after an objective change.
        It is backed by the limit.home.offset setting.

        Args:
            datum: Value of datum.
            unit: Units of datum.
        """
        request = dto.ObjectiveChangerSetRequest(
            interface_id=self.turret.connection.interface_id,
            turret_address=self.turret.device_address,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            value=datum,
            unit=unit,
        )
        call("objective_changer/set_datum", request)

    async def set_focus_datum_async(
            self,
            datum: float,
            unit: LengthUnits = Units.NATIVE
    ) -> None:
        """
        Sets the focus datum.
        The focus datum is the position that the focus stage moves to after an objective change.
        It is backed by the limit.home.offset setting.

        Args:
            datum: Value of datum.
            unit: Units of datum.
        """
        request = dto.ObjectiveChangerSetRequest(
            interface_id=self.turret.connection.interface_id,
            turret_address=self.turret.device_address,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            value=datum,
            unit=unit,
        )
        await call_async("objective_changer/set_datum", request)

    def __verify_is_changer(
            self
    ) -> None:
        """
        Checks if this is a objective changer and throws an error if it is not.
        """
        request = dto.ObjectiveChangerRequest(
            interface_id=self.turret.connection.interface_id,
            turret_address=self.turret.device_address,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
        )
        call_sync("objective_changer/verify", request)

    def __repr__(
            self
    ) -> str:
        """
        Returns a string that represents the device.

        Returns:
            A string that represents the device.
        """
        request = dto.AxisToStringRequest(
            interface_id=self.turret.connection.interface_id,
            device=self.turret.device_address,
        )
        response = call_sync(
            "device/device_to_string",
            request,
            dto.StringResponse.from_binary)
        return response.value



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import Generator, Any, Optional, List
import asyncio

from ..call import call, call_async, call_sync
from ..exceptions.motion_lib_exception import MotionLibException

from ..dto import requests as dto
from ..dto.microscopy.wdi_autofocus_provider_status import WdiAutofocusProviderStatus


class WdiAutofocusProvider:
    """
    Class representing access to WDI Autofocus connection.
    """

    DEFAULT_TCP_PORT = 27
    """
    Default port number for TCP connections to WDI autofocus.
    """

    @property
    def provider_id(self) -> int:
        """
        The ID identifies the autofocus with the underlying library.
        """
        return self._provider_id

    def __init__(self, provider_id: int):
        self._provider_id = provider_id

    @staticmethod
    def open_tcp(
            host_name: str,
            port: int = DEFAULT_TCP_PORT
    ) -> 'WdiAutofocusProvider':
        """
        Opens a TCP connection to WDI autofocus.

        Args:
            host_name: Hostname or IP address.
            port: Optional port number (defaults to 27).

        Returns:
            An object representing the autofocus connection.
        """
        request = dto.OpenInterfaceRequest(
            interface_type=dto.InterfaceType.TCP,
            host_name=host_name,
            port=port,
        )
        response = call(
            "wdi/open",
            request,
            dto.IntResponse.from_binary)
        return WdiAutofocusProvider(response.value)

    @staticmethod
    def open_tcp_async(
            host_name: str,
            port: int = DEFAULT_TCP_PORT
    ) -> 'AsyncWdiAutofocusOpener':
        """
        Opens a TCP connection to WDI autofocus.

        Args:
            host_name: Hostname or IP address.
            port: Optional port number (defaults to 27).

        Returns:
            An object representing the autofocus connection.
        """
        request = dto.OpenInterfaceRequest(
            interface_type=dto.InterfaceType.TCP,
            host_name=host_name,
            port=port,
        )
        return AsyncWdiAutofocusOpener(request)

    def close(
            self
    ) -> None:
        """
        Close the connection.
        """
        request = dto.InterfaceEmptyRequest(
            interface_id=self.provider_id,
        )
        call("wdi/close", request)

    async def close_async(
            self
    ) -> None:
        """
        Close the connection.
        """
        request = dto.InterfaceEmptyRequest(
            interface_id=self.provider_id,
        )
        await call_async("wdi/close", request)

    def generic_read(
            self,
            register_id: int,
            size: int,
            count: int = 1,
            offset: int = 0,
            register_bank: str = "U"
    ) -> List[int]:
        """
        Generic read operation.

        Args:
            register_id: Register address to read from.
            size: Data size to read. Valid values are (1,2,4). Determines the data type (Byte, Word, DWord).
            count: Number of values to read (defaults to 1).
            offset: Offset within the register (defaults to 0).
            register_bank: Register bank letter (defaults to U for user bank).

        Returns:
            Array of integers read from the device.
        """
        request = dto.WdiGenericRequest(
            interface_id=self.provider_id,
            register_id=register_id,
            size=size,
            count=count,
            offset=offset,
            register_bank=register_bank,
        )
        response = call(
            "wdi/read",
            request,
            dto.IntArrayResponse.from_binary)
        return response.values

    async def generic_read_async(
            self,
            register_id: int,
            size: int,
            count: int = 1,
            offset: int = 0,
            register_bank: str = "U"
    ) -> List[int]:
        """
        Generic read operation.

        Args:
            register_id: Register address to read from.
            size: Data size to read. Valid values are (1,2,4). Determines the data type (Byte, Word, DWord).
            count: Number of values to read (defaults to 1).
            offset: Offset within the register (defaults to 0).
            register_bank: Register bank letter (defaults to U for user bank).

        Returns:
            Array of integers read from the device.
        """
        request = dto.WdiGenericRequest(
            interface_id=self.provider_id,
            register_id=register_id,
            size=size,
            count=count,
            offset=offset,
            register_bank=register_bank,
        )
        response = await call_async(
            "wdi/read",
            request,
            dto.IntArrayResponse.from_binary)
        return response.values

    def generic_write(
            self,
            register_id: int,
            size: int = 0,
            data: List[int] = [],
            offset: int = 0,
            register_bank: str = "U"
    ) -> None:
        """
        Generic write operation.

        Args:
            register_id: Register address to read from.
            size: Data size to write. Valid values are (0,1,2,4). Determines the data type (Nil, Byte, Word, DWord).
            data: Array of values to write to the register. Empty array is allowed.
            offset: Offset within the register (defaults to 0).
            register_bank: Register bank letter (defaults to U for user bank).
        """
        request = dto.WdiGenericRequest(
            interface_id=self.provider_id,
            register_id=register_id,
            size=size,
            data=data,
            offset=offset,
            register_bank=register_bank,
        )
        call("wdi/write", request)

    async def generic_write_async(
            self,
            register_id: int,
            size: int = 0,
            data: List[int] = [],
            offset: int = 0,
            register_bank: str = "U"
    ) -> None:
        """
        Generic write operation.

        Args:
            register_id: Register address to read from.
            size: Data size to write. Valid values are (0,1,2,4). Determines the data type (Nil, Byte, Word, DWord).
            data: Array of values to write to the register. Empty array is allowed.
            offset: Offset within the register (defaults to 0).
            register_bank: Register bank letter (defaults to U for user bank).
        """
        request = dto.WdiGenericRequest(
            interface_id=self.provider_id,
            register_id=register_id,
            size=size,
            data=data,
            offset=offset,
            register_bank=register_bank,
        )
        await call_async("wdi/write", request)

    def enable_laser(
            self
    ) -> None:
        """
        Enables the laser.
        """
        request = dto.WdiGenericRequest(
            interface_id=self.provider_id,
            register_id=1,
        )
        call("wdi/write", request)

    async def enable_laser_async(
            self
    ) -> None:
        """
        Enables the laser.
        """
        request = dto.WdiGenericRequest(
            interface_id=self.provider_id,
            register_id=1,
        )
        await call_async("wdi/write", request)

    def disable_laser(
            self
    ) -> None:
        """
        Disables the laser.
        """
        request = dto.WdiGenericRequest(
            interface_id=self.provider_id,
            register_id=2,
        )
        call("wdi/write", request)

    async def disable_laser_async(
            self
    ) -> None:
        """
        Disables the laser.
        """
        request = dto.WdiGenericRequest(
            interface_id=self.provider_id,
            register_id=2,
        )
        await call_async("wdi/write", request)

    def get_status(
            self
    ) -> WdiAutofocusProviderStatus:
        """
        Gets the status of the autofocus.

        Returns:
            The status of the autofocus.
        """
        request = dto.InterfaceEmptyRequest(
            interface_id=self.provider_id,
        )
        response = call(
            "wdi/get_status",
            request,
            dto.WdiGetStatusResponse.from_binary)
        return response.status

    async def get_status_async(
            self
    ) -> WdiAutofocusProviderStatus:
        """
        Gets the status of the autofocus.

        Returns:
            The status of the autofocus.
        """
        request = dto.InterfaceEmptyRequest(
            interface_id=self.provider_id,
        )
        response = await call_async(
            "wdi/get_status",
            request,
            dto.WdiGetStatusResponse.from_binary)
        return response.status

    def __repr__(
            self
    ) -> str:
        """
        Returns a string that represents the autofocus connection.

        Returns:
            A string that represents the connection.
        """
        request = dto.InterfaceEmptyRequest(
            interface_id=self.provider_id,
        )
        response = call_sync(
            "wdi/to_string",
            request,
            dto.StringResponse.from_binary)
        return response.value

    def __enter__(self) -> 'WdiAutofocusProvider':
        """ __enter__ """
        return self

    def __exit__(self, _type: Any, _value: Any, _traceback: Any) -> None:
        """ __exit__ """
        self.close()


class AsyncWdiAutofocusOpener:
    '''Provides a connection in an asynchronous context using `await` or `async with`'''
    def __init__(self, request: dto.OpenInterfaceRequest) -> None:
        self._request = request
        self._resource: Optional[WdiAutofocusProvider] = None

    async def _create_resource(self) -> WdiAutofocusProvider:
        task = asyncio.ensure_future(call_async(
            "wdi/open",
            self._request,
            dto.IntResponse.from_binary))

        try:
            response = await asyncio.shield(task)
        except asyncio.CancelledError:
            async def cancel() -> None:
                try:
                    response = await task
                    await WdiAutofocusProvider(response.value).close_async()
                except MotionLibException:
                    pass

            asyncio.ensure_future(cancel())
            raise

        return WdiAutofocusProvider(response.value)

    def __await__(self) -> Generator[Any, None, 'WdiAutofocusProvider']:
        return self._create_resource().__await__()

    async def __aenter__(self) -> 'WdiAutofocusProvider':
        self._resource = await self._create_resource()
        return self._resource

    async def __aexit__(self, exc_type: Any, exc: Any, trace: Any) -> None:
        if self._resource is not None:
            await self._resource.close_async()



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import List
from ..call import call, call_sync, call_async

from ..dto import requests as dto
from ..ascii.device import Device
from .process import Process
from ..ascii.connection import Connection


class ProcessController:
    """
    Use to manage a process controller.
    Requires at least Firmware 7.35.
    """

    @property
    def device(self) -> Device:
        """
        The base device of this process controller.
        """
        return self._device

    def __init__(self, device: Device):
        """
        Creates instance of `ProcessController` of the given device.
        If the device is identified, this constructor will ensure it is a process controller.
        """
        self._device = device
        self.__verify_is_process_controller()

    @staticmethod
    def detect(
            connection: Connection,
            identify: bool = True
    ) -> List['ProcessController']:
        """
        Detects the process controllers on the connection.

        Args:
            connection: The connection to detect process controllers on.
            identify: If the Process Controllers should be identified upon detection.

        Returns:
            A list of all `ProcessController`s on the connection.
        """
        request = dto.DeviceDetectRequest(
            type=dto.DeviceType.PROCESS_CONTROLLER,
            interface_id=connection.interface_id,
            identify_devices=identify,
        )
        response = call(
            "device/detect",
            request,
            dto.DeviceDetectResponse.from_binary)
        return [ProcessController(connection.get_device(device)) for device in response.devices]

    @staticmethod
    async def detect_async(
            connection: Connection,
            identify: bool = True
    ) -> List['ProcessController']:
        """
        Detects the process controllers on the connection.

        Args:
            connection: The connection to detect process controllers on.
            identify: If the Process Controllers should be identified upon detection.

        Returns:
            A list of all `ProcessController`s on the connection.
        """
        request = dto.DeviceDetectRequest(
            type=dto.DeviceType.PROCESS_CONTROLLER,
            interface_id=connection.interface_id,
            identify_devices=identify,
        )
        response = await call_async(
            "device/detect",
            request,
            dto.DeviceDetectResponse.from_binary)
        return [ProcessController(connection.get_device(device)) for device in response.devices]

    def get_process(
            self,
            process_number: int
    ) -> Process:
        """
        Gets an Process class instance which allows you to control a particular voltage source.
        Axes are numbered from 1.

        Args:
            process_number: Number of process to control.

        Returns:
            Process instance.
        """
        if process_number <= 0:
            raise ValueError('Invalid value; processes are numbered from 1.')

        return Process(self, process_number)

    def __verify_is_process_controller(
            self
    ) -> None:
        """
        Checks if this is a process controller or some other type of device and throws an error if it is not.
        """
        request = dto.DeviceEmptyRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
        )
        call_sync("process_controller/verify", request)

    def __repr__(
            self
    ) -> str:
        """
        Returns a string that represents the device.

        Returns:
            A string that represents the device.
        """
        request = dto.AxisToStringRequest(
            interface_id=self.device.connection.interface_id,
            device=self.device.device_address,
        )
        response = call_sync(
            "device/device_to_string",
            request,
            dto.StringResponse.from_binary)
        return response.value
